import numpy as np
import pytest
import scipy

from nacl.sparseutils import kron_product_by_line
from bbf.bspline import BSpline


def test_wont_work():
    with pytest.raises(ValueError) as err:
        kron_product_by_line(
            np.asarray([1., 2.]),
            np.asarray([1., 2.]))
    assert "both arrays should be 2-dimensional" in str(err)

    with pytest.raises(ValueError) as err:
        kron_product_by_line(
            np.asarray([1., 2.]).reshape(1, 2),
            np.asarray([1., 2.]))
    assert "both arrays should be 2-dimensional" in str(err)

    with pytest.raises(ValueError) as err:
        kron_product_by_line(
            np.asarray([1., 2.]).reshape(1, 2),
            np.asarray([1., 2., 3., 4.]).reshape(2, 2))
    assert "number of rows for both arrays should be equal" in str(err)

    A = scipy.sparse.csr_matrix(
        np.asarray([1., 2., 3., 4.]).reshape((2, 2)))
    B = np.asarray([0., 5., 6., 7.]).reshape((2, 2))

    with pytest.raises(ValueError) as err:
        kron_product_by_line(B, B)
    assert "must be a sparse array" in str(err)

    with pytest.raises(ValueError) as err:
        kron_product_by_line(A, A)
    assert "must be a dense array" in str(err)


def test_kron_product_by_line_1():
    A = scipy.sparse.csr_matrix(
        np.asarray([1., 2., 3., 4.]).reshape((2, 2)))
    B = np.asarray([0., 5., 6., 7.]).reshape((2, 2))

    C = kron_product_by_line(A, B)
    assert C.shape == (2, 4)
    assert np.allclose(
        C.toarray().flatten(),
        np.asarray([0, 5, 0, 10, 18, 21, 24, 28], dtype=np.float64))


def test_kron_product_by_line_2():
    basis = BSpline(np.linspace(-10., 10., 3))
    J = basis.eval(np.asarray([0.1, 0.3, 0.2]))
    assert J.shape == (3, 5)

    F = np.asarray(
        [0.2, 0.1, 0.3, 0, 0, 0.5, 1, 1.2, 0, 0.1, 0, 0, 0, 0, 0]
    ).reshape((3, 5))

    K = kron_product_by_line(J, F)
    assert K.shape == (3, 25)
    assert np.allclose(K.data, np.asarray(
        [3.23433000e-02, 1.61716500e-02, 4.85149500e-02, 1.33313433e-01,
         6.66567167e-02, 1.99970150e-01, 3.43432333e-02, 1.71716167e-02,
         5.15148500e-02, 3.33333333e-08, 1.66666667e-08, 5.00000000e-08,
         7.60560833e-02, 1.52112167e-01, 1.82534600e-01, 1.52112167e-02,
         3.32890083e-01, 6.65780167e-01, 7.98936200e-01, 6.65780167e-02,
         9.10515833e-02, 1.82103167e-01, 2.18523800e-01, 1.82103167e-02,
         2.25000000e-06, 4.50000000e-06, 5.40000000e-06, 4.50000000e-07]))
    assert [5, 6, 7, 10, 11, 12, 15, 16, 17, 20, 21, 22, 5, 6, 7, 9, 10,
            11, 12, 14, 15, 16, 17, 19, 20, 21, 22, 24] == list(K.col)
    assert [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 1, 1, 1] == list(K.row)



import numpy as np
from scipy import sparse


class Constraint:
    """Generic contraint
    """

    def __init__(self, model, rhs):  # , pars_struct=None):
        """
        """
        self.model = model
        self.rhs = rhs
        # if pars_struct:
        #     self.pars_struct = pars_struct
        # else:
        #     self.pars_struct = self.model.pars.copy()
        #     self.pars_struct.release()

    def __call__(self, p=None, deriv=False):
        pass


class LinearConstraint(Constraint):
    """Generic linear contraints
    """
    def __init__(self, model, rhs):
        """
        """
        super().__init__(model, rhs)
        self.h_matrix = None
        self.rhs = rhs

    def init_h_matrix(self, pars):
        raise NotImplementedError()

    def init_pars(self, pars):
        """
        """
        self.h_matrix = self.init_h_matrix(pars)

    def __call__(self, pars, deriv=False):
        """evaluate the constraint
        """
        if self.h_matrix is None:
            self.h_matrix = self.init_h_matrix(pars)

        cons = self.h_matrix @ pars.full - self.rhs
        cons = float(cons)
        if not deriv:
            return cons
        return cons, self.h_matrix, None


class ConstraintSet:
    """Combine a series of constraints (linear or not)

    This class combines a set of constraints and produces a (quadratic)
    penality, added to the Log Likelihood. Compute the gradient and the hessian
    of this penality if required.
    """

    def __init__(self, constraints, mu=1.E10):
        """constructor
        """
        # self.model = model
        self.constraints = constraints
        self.mu = mu

    def init_pars(self, pars):
        """
        """
        for c in self.constraints:
            c.init_pars(pars)

    def __call__(self, pars, deriv=False):
        """evaluate the penality
        """
        npars = len(pars.full)

        pen = 0.
        # if no derivatives specified, return the sum of the quadratic
        # penalities associated with each constraint
        if not deriv:
            for cons in self.constraints:
                pen += cons(pars, deriv=False)**2
            return self.mu * float(pen)

        # otherwise, compute and return the gradient and hessian
        # along with the quadratic penality
        grad = coo_matrix(([], ([], [])), shape=(1,npars))
        hess = coo_matrix(([], ([],[])), shape=(npars,npars))
        for cons in self.constraints:
            # p=None, because self.model.pars was just updated
            c, dc, d2c = cons(pars, deriv=True)
            pen  += c**2
            # we have restored the true grad convention (-2 -> +2)
            grad += +2. * float(c) * dc
            hess += +2. * dc.T.dot(dc)
            if d2c is not None:
                hess += 2. * c * d2c

        # fixed parameters ?
        idx = pars.indexof() >= 0
        pen = float(pen)
        grad = np.array(grad[:,idx].todense()).squeeze()
        hess = hess[:,idx][idx,:]

        return self.mu * pen, self.mu * grad, self.mu * hess

    def get_rhs(self):
        return np.array([c.rhs for c in self.constraints])





class MaxLcAtPhase(LinearConstraint):

    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase

    def init_h_matrix(self, pars):
        J = self.model.basis.eval(np.array([self.phase])).tocoo()
        pp = pars.copy()
        pp.release()
        N = len(J.col)
        i = np.full(N, 0)
        j = pp['theta'].indexof(J.col)
        # idx = j>=0
        npars = len(pp.full)
        J = sparse.coo_matrix((J.data, (i, j)), shape=(1, npars))
        return J

    # def max_lc_at_phase(model, rhs, phase=0.):
    #     """
    #     """
    #     J =  model.basis.eval(np.array([phase])).toarray()
    #     return LinearConstraint(J, rhs)

class DMaxLcAtPhase(LinearConstraint):

    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase

    def init_h_matrix(self, pars):
        J = self.model.basis.deriv(np.array([self.phase])).tocoo()
        pp = pars.copy()
        pp.release()
        N = len(J.col)
        i = np.full(N, 0)
        j = pp['theta'].indexof(J.col)
        npars = len(pp.full)
        J = sparse.coo_matrix((J.data, (i, j)), shape=(1, npars))
        return J

class MeanStretch(LinearConstraint):
    # Not the constraint i want but worth testing
    def __init__(self, model, rhs):
        super().__init__(model, rhs)
        
    def init_h_matrix(self, pars):
        nsn = len(pars['x0'].full)
        npars = len(pars.full)
        j = pars['stretch'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = np.full(nsn, 1./nsn)
        idx = j >= 0
        J = sparse.coo_matrix((v[idx], (i[idx], j[idx])), shape=(1, npars))

        return J

class MeanColor(LinearConstraint):

    def __init__(self, model, rhs):
        super().__init__(model, rhs)

    def init_h_matrix(self, pars):
        nsn = len(pars['x0'].full)
        npars = len(pars.full)
        j = pars['color'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = np.full(nsn, 1./nsn)
        idx = j >= 0
        J = sparse.coo_matrix((v[idx], (i[idx], j[idx])), shape=(1, npars))

        return J

def cons(model, mu=1.E6, s=False):
    """
    """
    #    m0 = max_lc_at_phase(model, 0., phase=0.)
    #    dm0 = dmax_lc_at_phase(model, 0., phase=0.)
    m0 = MaxLcAtPhase(model, 1., 0.)
    dm0 = DMaxLcAtPhase(model, 0., 0.)
    if not s:
        return ConstraintSet([m0, dm0], mu=mu)
    else:
        s0 = MeanStretch(model, 0.)
        c0 = MeanColor(model, 0.)
        return ConstraintSet([m0, dm0, s0, c0], mu=mu)



r"""This module implements the constraints that are needed to train a
nacl.model.SALT2Like model on a dataset.

At variance with the original SALT2 work and with what is described in Guy
Augarde's thesis, we have made a special effort to implement constraints which
are linear.

The contraints are implemented as quadratic penalties that are added to the
:math:`\chi^2`. These penalities are typically of the form:

.. math ::
    (f(\theta) - \alpha)^2

where :math:`f(\theta)` is a function of the parameters, and :math:`\alpha`
is a number.

Since there are several constraints, it is convenient to express them in vector
form:

.. math ::
    (\vec{F}(\theta) - \vec{\alpha})^T \cdot (\vec{F}(\theta) - \vec{\alpha})
"""

import logging
import numpy as np
from scipy.sparse import coo_matrix, dok_matrix


class AllParametersFixedError(Exception): pass


def solve_constraints(cons, pars):
    """
    Solve the linearized constraints for the given parameters.

    This function linearizes the constraints provided by `cons` and solves
    for the parameter adjustments needed to satisfy these constraints.

    Parameters
    ----------
    cons : Constraints
        The constraints object containing the linearized constraints and the right-hand side values.
    pars : Parameters
        The parameters object containing the current values of the parameters.

    Returns
    -------
    numpy.array
        The adjusted free parameters that satisfy the constraints.

    Notes
    -----
    The function uses QR decomposition to solve the linear system of equations formed by the constraints.
    """
    H = cons.get_linearized_constraints(pars.free)
    rhs = cons.get_rhs()
    Q,R = np.linalg.qr(H.T.todense())
    dx = np.linalg.solve(R.T, rhs - H @ pars.full)
    dx = np.array(Q.dot(dx)).squeeze()

    pp = pars.copy()
    pp.full[:] = dx

    return pp.free


def ab_flux_at_10Mpc(Mb=-19.5):
    """
    Calculate the AB flux at a distance of 10 Megaparsecs.

    Parameters
    ----------
    Mb : float, optional
        The absolute magnitude in the B-band (default is -19.5).

    Notes
    -----
    The formula used is `10**(-0.4 * (30 + Mb))`, where 30 is the distance modulus
    for 10 Megaparsecs.
    """
    return 10**(-0.4 * (30+Mb))


class Constraint:
    """
    Represents a constraint in the NaCl optimization problem.

    This class is used to define and apply constraints to the parameters
    of a model during optimization.
    """
    def __init__(self, model, rhs):
        """
        Initialize the Constraint class.

        Parameters
        ----------
        model : object
            The model to which the constraints are applied.
        rhs : numpy.array
            The right-hand side of the constraint equations.
        """
        self.model = model
        self.rhs = rhs

    def __call__(self, p=None, deriv=False):
        """
        Evaluate the constraint function or its derivative.

        Parameters
        ----------
        p : numpy.array, optional
            The parameters at which to evaluate the constraint (default is None).
        deriv : bool, optional
            If True, return the derivative of the constraint (default is False).

        Returns
        -------
        None
            This method should be implemented to return the constraint value or its derivative.
        """
        pass


class LinearConstraint(Constraint):
    """
    Generic linear constraints for a model.

    This class represents linear constraints applied to a model's parameters,
    ensuring that certain conditions are met during optimization.

    Parameters
    ----------
    model : object
        The model to which the constraints are applied.
    rhs : numpy.array
        The right-hand side of the constraint equations.
    """
    def __init__(self, model, rhs):
        """
        Initialize the LinearConstraint class.

        Parameters
        ----------
        model : object
            The model to which the constraints are applied.
        rhs : numpy.array
            The right-hand side of the constraint equations.
        """
        super().__init__(model, rhs)
        self.h_matrix = None
        self.rhs = rhs

    def init_h_matrix(self, pars):
        """
        Initialize the H matrix based on the parameters.

        This method should be implemented by subclasses.

        Parameters
        ----------
        pars : object
            The parameters used to initialize the H matrix.

        Raises
        ------
        NotImplementedError
            If the method is not implemented by a subclass.
        """
        raise NotImplementedError()

    def init_pars(self, pars):
        """
        Initialize the constraint with the given parameters.

        Parameters
        ----------
        pars : object
            The parameters used to initialize the constraint.
        """
        self.h_matrix = self.init_h_matrix(pars)

    def __call__(self, pars, deriv=False):
        """
        Evaluate the constraint and optionally its derivatives.

        Parameters
        ----------
        pars : object
            The parameters at which to evaluate the constraint.
        deriv : bool, optional
            If True, return the derivative of the constraint (default is False).

        Returns
        -------
        float
            The value of the constraint.
        tuple
            If deriv is True, returns a tuple containing the constraint value, the H matrix, and None.
        """
        if self.h_matrix is None:
            self.h_matrix = self.init_h_matrix(pars)

        cons = self.h_matrix @ pars.full - self.rhs
        cons = float(cons)
        if not deriv:
            return cons
        return cons, self.h_matrix, None


class ConstraintSet:
    """Combine a series of constraints (linear or not)

    This class combines a set of constraints and produces a (quadratic)
    penality, added to the Log Likelihood. Compute the gradient and the hessian
    of this penality if required.
    """

    def __init__(self, constraints, mu=1.E10):
        """constructor
        """
        # self.model = model
        self.constraints = constraints
        self.mu = mu

    def init_pars(self, pars):
        """
        """
        for c in self.constraints:
            c.init_pars(pars)

    def __call__(self, pars, deriv=False):
        """evaluate the penality
        """
        npars = len(pars.full)

        pen = 0.
        # if no derivatives specified, return the sum of the quadratic
        # penalities associated with each constraint
        if not deriv:
            for cons in self.constraints:
                pen += cons(pars, deriv=False)**2
            return self.mu * float(pen)

        # otherwise, compute and return the gradient and hessian
        # along with the quadratic penality
        grad = coo_matrix(([], ([], [])), shape=(1,npars))
        hess = coo_matrix(([], ([],[])), shape=(npars,npars))
        for cons in self.constraints:
            # p=None, because self.model.pars was just updated
            c, dc, d2c = cons(pars, deriv=True)
            pen  += c**2
            # we have restored the true grad convention (-2 -> +2)
            grad += +2. * float(c) * dc
            hess += +2. * dc.T.dot(dc)
            if d2c is not None:
                hess += 2. * c * d2c

        # fixed parameters ?
        idx = pars.indexof() >= 0
        pen = float(pen)
        grad = np.array(grad[:,idx].todense()).squeeze()
        hess = hess[:,idx][idx,:]

        return self.mu * pen, self.mu * grad, self.mu * hess

    def get_rhs(self):
        return np.array([c.rhs for c in self.constraints])


class int_M0_at_phase_cons(LinearConstraint):
    """constraint on the integral of the M0 surface at peak

    This function builds a linear constraint on the M0 parameters.

    .. note:: at this stage, the constraints are a function of the *full*
       parameter vector, not just the free parameters.
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase

    def init_h_matrix(self, pars):
        J_phase = self.model.basis.by.eval(np.array([self.phase])).toarray()
        pp = pars.copy()
        pp.release()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_phase, gram_dot_filter).ravel())
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        i = np.full(len(C.col), 0)
        j = pp['M0'].indexof(C.col)
        M = coo_matrix((C.data, (i, j)), shape=(1, npars))
        M.data *= (self.model.norm / self.model.int_ab_spec)
        return M


class int_dM0_at_phase_cons(LinearConstraint):
    """constraint on the integral of the phase derivatives of M0 at peak
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase
    def init_h_matrix(self, pars):
        J_dphase = self.model.basis.by.deriv(np.array([self.phase])).toarray()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_dphase, gram_dot_filter).ravel())
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['M0'].indexof(C.col)
        i = np.full(len(C.col), 0)
        v = C.data # / self.model.int_M0_phase_0 # (1.E5 * self.model.int_ab_spec)
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                        shape=(1, npars))
        M.data *= (self.model.norm)
        return M


class int_M1_at_phase_cons(LinearConstraint):
    """constraint on the integral of the M1 surface at peak
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase
    def init_h_matrix(self, pars):
        J_phase = self.model.basis.by.eval(np.array([self.phase])).toarray()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_phase, gram_dot_filter).ravel())
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['M1'].indexof(C.col)
        i = np.full(len(C.col), 0)
        v = C.data   #/ model.int_M0_phase_0
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                        shape=(1, npars))
        M.data *= self.model.norm
        return M


class int_dM1_at_phase_cons(LinearConstraint):
    """constraint on the integral of the phase derivatives of M1 at peak
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase
    def init_h_matrix(self, pars):
        J_dphase = self.model.basis.by.deriv(np.array([self.phase])).toarray()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_dphase, gram_dot_filter).ravel())
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['M1'].indexof(C.col)
        i = np.full(len(C.col), 0)
        v = C.data  # / model.int_M0_phase_0
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                        shape=(1, npars))
        M.data *= self.model.norm
        return M


class mean_col_cons(LinearConstraint):
    """constraint on the mean of the color parameters
    """
    def __init__(self, model, rhs):
        super().__init__(model, rhs)
    def init_h_matrix(self, pars):
        nsn = len(pars['X0'].full)
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['c'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = np.full(nsn, 1./nsn)
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])), shape=(1, npars))
        return M


class mean_x1_cons(LinearConstraint):
    """constraint on the mean of the x1 parameters
    """
    def __init__(self, model, rhs):
        super().__init__(model, rhs)
    def init_h_matrix(self, pars):
        nsn = len(pars['X0'].full)
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['X1'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = np.full(nsn, 1./nsn)
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                            shape=(1, npars))
        return M


class x1_var_cons(Constraint):

    def __init__(self, model, rhs):
        self.model = model
        self.rhs = rhs

    def __call__(self, pars, deriv=False):
        """
        """
        # if p is not None:
        #     self.model.pars.free = p.free

        # CHECK: do we need all_pars_released, or just self.model.pars ?
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full) # len(self.model.all_pars_released.full)
        nsn = len(pp['X0'].full)

        # constraint function: h(X1) = \sum_i X1**2
        cons = (pp['X1'].full**2).sum() / nsn  # pp or pars ?
        if not deriv:
            return cons

        # first derivatives of h
        pars.full[:] = 0.
        j = pars['X1'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = pars['X1'].full
        idx = j >= 0
        J = coo_matrix((v[idx], (i[idx], j[idx])), shape=(1, npars)) # was +1

        # second derivatives of h
        i = pars['X1'].indexof(np.arange(nsn))
        v = np.full(nsn, 2./nsn)
        idx = j >= 0
        H = coo_matrix((v[idx], (i[idx], i[idx])), shape=(npars, npars))

        return cons, J, H


def salt2like_linear_constraints(model, mu=1.E6, Mb=-19.5, dm15=1.): # was 0.96
    """
    """
    m0_0 = int_M0_at_phase_cons(model, 14381300.77605067, phase=0.) # ab_flux_at_10Mpc(Mb=Mb),
    dm0_0 = int_dM0_at_phase_cons(model, 0., phase=0.)
    m1_0 = int_M1_at_phase_cons(model, 0., phase=0.)
    m1_15 = int_M1_at_phase_cons(model, dm15, phase=15.)  # rhs was 0.96
    # dm1 = int_dM1_at_phase_cons(model, 0.)
    col = mean_col_cons(model, 0.)
    x1  = mean_x1_cons(model, 0.)
    # x1_var = x1_var_cons(model, 1.)

    return ConstraintSet([m0_0, dm0_0, m1_0, m1_15, col, x1], mu=mu)


def salt2like_classical_constraints(model, mu=1.E6, Mb=-19.5):
    """
    """
    # m0_0 = int_M0_at_phase_cons(model, ab_flux_at_10Mpc(Mb=Mb), phase=0.)
    m0_0 = int_M0_at_phase_cons(model, 14381300.77605067, phase=0.)
    dm0_0 = int_dM0_at_phase_cons(model, 0., phase=0.)
    m1_0 = int_M1_at_phase_cons(model, 0., phase=0.)
    dm1 = int_dM1_at_phase_cons(model, 0.)
    col = mean_col_cons(model, 0.)
    x1  = mean_x1_cons(model, 0.)
    x1_var = x1_var_cons(model, 1.)

    return ConstraintSet(
        model,
        [m0_0, dm0_0, m1_0, dm1, col, x1, x1_var],
        mu=mu)



"""Pure python reimplementation of the original salt2 model
"""


import logging
import numpy as np
import scipy.sparse
import sncosmo

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt

from bbf import bspline, SNFilterSet, SNMagSys
from saltworks import FitParameters

from .colorlaw import ColorLaw
from .lightcurves import LightcurveEvalUnit
from .spectra import CompressedSpectrumEvalUnit, CompressedSpectroPhotoEvalUnit #  SpectrumRecalibrationPolynomials



class SALT2Like(object):
    r"""A re-implementation of the SALT2 model

    SALT2 is an empirical SN spectrophotometric model. This class provides a
    pure python re-implementation of SALT2 with a number of improvements.

    The SALT2 parametrization is defined as follows. In the SN restframe the
    absolute spectral density :math:`S(\lambda)` is:

    .. math::
        S(\lambda, \mathrm{p}) = X_0 \times \left[M_0(\lambda, \mathrm{p}) + X_1
           \times M_1(\lambda, \mathrm{p})\right]\ 10^{0.4 c CL(\lambda)}

    where :math:`\mathrm{p}` is the SN phase, i.e. the restframe time since SN
    peak luminosity:

    .. math::
        \mathrm{ph} = \frac{t_{MJD} - t_{max}}{1 + z}

    :math:`M_0`, :math:`M_1` are global surfaces describing the spectral
    evolution of the average SN and its principal variation, and
    :math:`CL(\lambda)` is a global "extinction correction" describing the color
    diversity of the SNIa family.

    :math:`(X_0, X_1, c)` are SN-dependent. :math:`X_0` is the amplitude of the
    "SN-frame B-band lightcurve" as inferred from the observer-frame fluxes.
    :math:`X_1` is the coordinate of the SN along :math:`M_1`. In practice, it
    quantifies the variations of the light curve width. :math:`c` is the SN
    color -- or more exactly, the SN color excess with respect to the average SN
    color.

    """

    def __init__(self, training_dataset,
                 phase_range=(-20., 50.), wl_range=(2000., 9000.),
                 basis_knots=[127, 20],
                 filter_wl_range=(2000., 11000.),
                 basis_filter_knots=1100,
                 wl_grid=None, phase_grid=None,
                 spectrum_recal_degree=3,
                 normalization_band_name='swope2::b',
                 disable_cache=False):
        """Constructor for the SALT2Like class

        Initializes the bases (phase, wavelength, filter), computes the Gram
        matrix, evaluates the color law on the effective wavelengths, and
        organizes the data into spectral/photometric computing units.

        Parameters
        ----------
        training_dataset : nacl.dataset.TrainingDataset
            The training dataset containing light curve and spectral data.
        phase_range : tuple of float, optional
            Nominal phase range of the model, by default (-20., 50.).
        wl_range : tuple of float, optional
            Nominal wavelength range of the model, by default (3000., 9000.).
        basis_knots : list of int, optional
            Number of wavelength and photometric knots, by default [127, 20].
        basis_filter_knots : int, optional
            Number of knots for the filter basis, by default, 900.
        wl_grid : array-like, optional
            Grid of wavelengths, to be used for basis construction.
        phase_grid : array-like, optional
            Grid of phases, to be used for basis construction.
        spectrum_recal_degree : int, optional
            Degree of the spectrum recalibration polynomials, by default 3.
        normalization_band_name : str, optional
            Name of the band to be used for model normalization, by default swope2::b
        disable_cache : bool, optional
            If True, disable the caching system, by default False.
        """
        self.training_dataset = training_dataset
        self.lc_data = training_dataset.lc_data
        self.phase_range = phase_range
        self.wl_range = wl_range
        self.filter_wl_range = filter_wl_range
        self.n_wl, self.n_ph = basis_knots[0],  basis_knots[1]
        self.basis_knots = basis_knots
        self.basis_filter_knots = basis_filter_knots
        self.spectrum_recal_degree = spectrum_recal_degree
        self.delta_phase = 0. # check: do not remember what this is
        # self.filterpath = self.training_dataset.filterpath
        if self.training_dataset.lc_data is not None:
            self.bands = list(self.training_dataset.transmissions.keys())
        self.normalization_band_name = normalization_band_name

        self.timing = []

        self.basis, self.gram, self.G2, self.L_eff = None, None, None, None
        self.val, self.jacobian_val, self.jacobian_i, self.jacobian_j = None, None, None, None
        self.polynome_color_law, self.jacobian_color_law = None, None


        # tests whether all data points are within the model range.
        # if not, print a warning
        valid = SALT2Like.flag_out_of_range_datapoints(training_dataset,
                                                       phase_range=phase_range,
                                                       wl_range=wl_range,
                                                       basis_knots=basis_knots,
                                                       basis_filter_knots=basis_filter_knots,
                                                       wl_grid=wl_grid, phase_grid=phase_grid,
                                                       compress=False)
        n_out = list(map(lambda x: 0 if x is None else (~x).sum(), valid))
        if n_out[0] > 0:
            logging.warning(f'{n_out[0]} lc data points outside model definition range')
        if n_out[1] > 0:
            logging.warning(f'{n_out[1]} spec data points outside model definition range')
        if n_out[2] > 0:
            logging.warning(f'{n_out[2]} spectrophotometric data points outside model definition range')

        # Filter database
        # TODO: we should be able to optimize this a little
        # (i.e. adapt the filter bases, to lower the size of the grams)
        self.filter_basis = bspline.BSpline(np.linspace(self.filter_wl_range[0],
                                                        self.filter_wl_range[1],
                                                        basis_filter_knots), order=4)
        self.filter_db = SNFilterSet(basis=self.filter_basis)
        # should be able to specify this as parameters
        self.filter_db.insert('swope2::b', z=0.)
        self.filter_db.insert('swope2::v', z=0.)
        self.color_law = ColorLaw()

        n_lc_meas, n_spec_meas, n_spectrophot_meas = \
            self.training_dataset.nb_meas(split_by_type=1)

        # TODO: where is this used ?
        # TODO: we may replace this by n_lc_meas > 0
        if self.training_dataset.lc_data is not None:
            tds = self.training_dataset
            ii = self.training_dataset.i_lc_first
            self.lambda_c = tds.lc_data.wavelength[ii] / (1+tds.lc_data.z[ii])
            self.lambda_c_red = self.color_law.reduce(self.lambda_c)

        # initialize model basis and model parameter vector
        # model component: bases, color law, and parameter vector
        self.basis = SALT2Like._init_bases(wl_grid, phase_grid,
                                           wl_range, phase_range, basis_knots)

        self.norm = self.normalization(pars=None, band_name=normalization_band_name,
                                       default_norm=1.01907246e-12)

        # grams
        self.init_grams_and_cc_grid()

        # and finally, prepare the computing units
        self.queue = self._init_eval_units()

        # cache system
        self.disable_cache = disable_cache
        self.cached = False
        self._cache = None

    @classmethod
    def flag_out_of_range_datapoints(cls, training_dataset,
                                     phase_range=(-20., 50.), wl_range=(2000., 9000.),
                                     basis_knots=[127, 20],
                                     basis_filter_knots=900,
                                     wl_grid=None, phase_grid=None,
                                     compress=False):
        """
        Flags data points in the training dataset that are out of the specified model phase and wavelength ranges.

        Parameters
        ----------
        training_dataset : TrainingDataset
            The training dataset containing lightcurves and/or spectral data
        phase_range : tuple of float, optional
            The valid range of phases (default is (-20., 50.)).
        wl_range : tuple of float, optional
            The valid range of wavelengths (default is (2000., 9000.)).
        basis_knots : list of int, optional
            The number of knots for the B-spline bases in wavelength and phase (default is [127, 20]).
        basis_filter_knots : int, optional
            The number of knots for the filter B-spline basis (default is 900).
        wl_grid : array-like, optional
            The wavelength grid for the model basis (default is None).
        phase_grid : array-like, optional
            The phase grid for the model basis (default is None).
        compress : bool, optional
            If True, compress the training dataset to exclude flagged data points (default is False).

        Returns
        -------
        tuple
            A tuple containing:
            - lc_valid (array-like or None): Boolean array indicating valid light curve data points.
            - spec_valid (array-like or None): Boolean array indicating valid spectral data points.
            - spectrophot_valid (array-like or None): Boolean array indicating valid spectrophotometric data points.

        Notes
        -----
        This method performs the following operations:
        1. Initializes model and filter bases using the provided grids and ranges.
        2. Calculates the valid phase and wavelength ranges from the model basis.
        3. Flags light curve data points that are outside the model's phase range.
        4. Flags spectral data points that are outside the model's phase and wavelength ranges.
        5. Flags spectrophotometric data points that are outside the model's phase and wavelength ranges.
        6. If `compress` is True, updates the training dataset to exclude flagged data points.

        Logging
        -------
        Logs the number of data points removed for each type (light curve, spectral, spectrophotometric).

        Examples
        --------
        >>> from nacl.dataset import TrainingDataset
        >>> from lemaitre import bandpasses
        >>> tds = TrainingDataset.read_parquet("data/test_datasets/test_datasets_blind.parquet")
        >>> lc_valid, spec_valid, spectrophot_valid = SALT2Like.flag_out_of_range_datapoints(tds, compress=True)
        """
        tds = training_dataset
        model_basis = cls._init_bases(wl_grid, phase_grid,
                                      wl_range, phase_range, basis_knots)
        filter_basis = bspline.BSpline(np.linspace(wl_range[0], wl_range[1],
                                                   basis_filter_knots), order=4)
        filter_db = SNFilterSet(basis=filter_basis)

        # model phase range, model wavelength range
        ph_min, ph_max = model_basis.by.grid.min(), model_basis.by.grid.max()
        wl_min, wl_max = model_basis.bx.grid.min(), model_basis.bx.grid.max()

        tmax = np.zeros(len(tds.sn_data))
        tmax[tds.sn_data.sn_index] = tds.sn_data.tmax[:]

        if tds.lc_data is not None:
            # select the light curves for which the filters
            # overlap with the model definition range
            valid = {}
            for lc in tds.lc_db:
                tr = filter_db.insert(lc.band, lc.z)
                valid[lc.lc_index] = tr.valid
            # tr_data = [filter_db.insert(lc.band, lc.z) for lc in tds.lc_db]
            # valid = np.array([tr.valid for tr in tr_data])
            lc_valid = np.array([valid[lci] for lci in tds.lc_data.lc_index])
            # select the light curve data points which are
            # in the model phase range
            restframe_phase = (tds.lc_data.mjd - tmax[tds.lc_data.sn_index]) / (1. + tds.lc_data.z)
            lc_valid &= ((restframe_phase >= ph_min) & (restframe_phase <= ph_max))
            tds.restframe_phase = restframe_phase
            # logging.info(f' --> {(~((restframe_phase >= ph_min) & (restframe_phase <= ph_max))).sum()}')
            n_out = (~lc_valid).sum()
            if n_out > 0:
                logging.warning(f'{n_out} lc data points outside the model definition range')
        else:
            lc_valid = None

        if tds.spec_data is not None:
            # select the spectra which are in the model phase range
            # tmax = tds.sn_data.tmax[tds.spec_data.sn_index]
            restframe_phase = (tds.spec_data.mjd - tmax[tds.spec_data.sn_index]) / (1. + tds.spec_data.z)
            spec_valid = (restframe_phase >= ph_min) & (restframe_phase <= ph_max)
            # select the spectral points which are in the model spectral range
            restframe_wl = tds.spec_data.wavelength / (1. + tds.spec_data.z)
            spec_valid &= (restframe_wl >= wl_min) & (restframe_wl <= wl_max)
            n_out = (~spec_valid).sum()
            if n_out > 0:
                logging.warning(f'{n_out} spec data points outside the model definition range')
        else:
            spec_valid = None

        if tds.spectrophotometric_data is not None:
            # same cut, for the spectrophotometric data
            # tmax = tds.sn_data.tmax[tds.spectrophotometric_data.sn_index]
            restframe_phase = (tds.spectrophotometric_data.mjd - tmax[tds.spectrophotometric_data.sn_index]) / (1. + tds.spectrophotometric_data.z)
            spectrophot_valid = (restframe_phase >= ph_min) & (restframe_phase <= ph_max)
            restframe_wl = tds.spectrophotometric_data.wavelength / (1. + tds.spectrophotometric_data.z)
            spectrophot_valid &= (restframe_wl >= wl_min) & (restframe_wl <= wl_max)
            n_out = (~spectrophot_valid).sum()
            if n_out > 0:
                logging.warning(f'{n_out} spectrophot data points outside the model definition range')
        else:
            spectrophot_valid = None

        if compress:
            if lc_valid is not None:
                tds.lc_data.valid &= lc_valid
                lc_valid = lc_valid[lc_valid]
            if spec_valid is not None:
                tds.spec_data.valid &= spec_valid
                spec_valid = spec_valid[spec_valid]
            if spectrophot_valid is not None:
                tds.spectrophotometric_data.valid &= spectrophot_valid
                spectrophot_valid = spectrophot_valid[spectrophot_valid]
            tds.compress()

        return (lc_valid, spec_valid, spectrophot_valid)

    def _get_from_cache(self, pars, jac=False):
        """
        Retrieve the cached model results and Jacobian matrix (if applicable).

        This method checks if the current parameter set matches the parameters
        stored in the cache. If they match, it retrieves the cached model results,
        and optionally, the Jacobian matrix.

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters
            The current set of fit parameters.
        jac : bool, optional
            Flag indicating whether the Jacobian matrix is needed. Defaults to False.

        Returns
        -------
        tuple or None
            - If `jac` is False and the cache is valid, returns the cached model results (numpy.array).
            - If `jac` is True and the cache is valid, returns a tuple containing the cached model results and Jacobian matrix (numpy.array, scipy.sparse.csr_matrix).
            - If the cache is invalid or not present, returns None.

        Notes
        -----
        The cache is validated by comparing the current parameters (`pars.full`) with the cached parameters.
        If the parameters do not match, the cache is considered invalid and None is returned.
        """

        if self.disable_cache or not hasattr(self, '_cache') or self._cache is None:
            return None

        p, v, J = self._cache
        if len(pars.full) != len(p.full):
            return None
        # TODO: check fixed pars
        if np.allclose(pars.full, p.full):
            if not jac:
                return v
            else:
                if J is None:
                    return None
                return v, J
        return None

    def  _save_to_cache(self, pars, v, J=None):
        """
        Save the current model results and Jacobian matrix to the cache.

        This method stores the provided model results and optionally the Jacobian
        matrix in the cache along with the current parameter set. This allows for
        quick retrieval of these values in future computations if the parameters
        have not changed.

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters
            The current set of fit parameters to be cached.
        v : numpy.array
            The model results to be cached.
        J : scipy.sparse.csr_matrix, optional
            The Jacobian matrix to be cached. Defaults to None.
        """
        if not self.disable_cache:
            self._cache = (pars.copy(), v, J)
        else:
            self._cache = None

    def get_gram_dot_filter(self, band_name=None):
        """Used by the integral constraints
        """
        key = band_name if band_name is not None else self.normalization_band_name
        coeffs = self.filter_db[key].tq
        return self.gram.dot(coeffs)

    def reduce(self, wl):
        return self.color_law.reduce(wl)

    def init_grams_and_cc_grid(self):
        r"""Compute the :math:`\Lambda^{\mathrm{eff}}_{\ell q}` matrix (see definition above) on which the color
        law is evaluated.

        .. math::
             \bar{\lambda}_{\ell q} = \frac{\int \lambda^2 B_\ell(\lambda) B_q(\lambda)
             d\lambda}{\int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda}

        Compute de grammian of order one, :math:`G` and two :math:`G2` of the model :

        .. math::
            G = \int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda \\
            G2 = \int \lambda^2 B_\ell(\lambda) B_q(\lambda) d\lambda


        .. note::
             this method will be moved to the main model class (since we may work
             with one single :math:`\Lambda^{\mathrm{eff}}` matrix in a near future).
        """
        self.gram = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=1)
        self.G2 = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=2)

        gram = self.gram.tocoo()
        gram2 = self.G2.tocoo()
        assert(~np.any(gram.row-gram2.row) and ~np.any(gram.col-gram2.col))
        l_eff = gram2.data / gram.data
        self.L_eff = scipy.sparse.coo_matrix((l_eff, (gram.row, gram.col)), shape=gram.shape)

    def get_struct(self):
        """return the structure of the fit parameter vector

        In practice, the fit parameter vector is larger than just the model
        parameters: it also contains the parameters of the error models (error
        snake, color scatter, calibration, see e.g. variancemodels.py). The
        instantiation of the final fit parameter vector cannot therefore be
        performed by the model itself. What the model can do, is to return the
        description of the fit parameter blocks it knows about.

        """
        nb_sne = self.training_dataset.nb_sne()
        # nb_spectra = self.training_dataset.nb_spectra()
        # nb_passbands = len(self.bands)
        # spec_recalibration_npars = self.recalibration_degree + 1
        # nb_lightcurves = self.training_dataset.nb_lcs(valid_only=False)
        d = [('X0',   nb_sne),
             ('X1',   nb_sne),
             ('c',  nb_sne),
             ('tmax', nb_sne),
             ('M0',   self.basis.bx.nj * self.basis.by.nj),
             ('M1',   self.basis.bx.nj * self.basis.by.nj),
             ('CL',  4)]
        if hasattr(self, 'recal_func'):
            d.extend(self.recal_func.get_struct())
            #        if self.training_dataset.spec_data is not None:
            #            d.append(('SpectrumRecalibration', spec_recalibration_npars.sum()))
        return d

    def init_pars(self, pars=None, model_name='salt2.4', version='2.4'):
        """
        Instantiate and initialize a fit parameter vector

        This method initializes the fit parameters for the SALT2 model. If no
        parameters are provided, it creates a new `FitParameters` object based
        on the model's structure. It then initializes the parameter vector. The
        SN-specific parameters are initialized from the `TrainingDataset` held
        by the model. The model-specific parameters (M0, M1, CL) are built from
        previous models (e.g. SALT2 or SALT3, retrieved from `sncosmo`)

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters, optional
            Existing fit parameters to initialize. If None, a new `FitParameters` object is created.
        model_name : str, optional
            The name of the model version to use for initialization, by default 'salt2.4'.
        version : str, optional
            The version of the model to use for initialization, by default '2.4'.

        Returns
        -------
        pars : nacl.fitparameters.FitParameters
            Initialized model parameters.

        Notes
        -----
        The model parameters are obtained from a projection of the SALT2/SALT3
        surfaces on the model basis. As the model basis may differ from the
        original SALT2/SALT3 bases, this initialization yields only an
        approximation of the original models.
        """
        if pars is None:
            pars = FitParameters(self.get_struct())

        self.init_model_pars(pars, model_name=model_name, version=version)
        self.init_from_training_dataset(pars)

        if 'SpectrumRecalibration' in pars._struct:
            self.recal_func.init_pars(pars)

        return pars

    def init_from_training_dataset(self, pars):
        """Load initial SN parameters from the training dataset.

        This method initializes the SN-specific parameters (X0, X1, c, tmax)
        from the `TrainingDataset` held by the model.

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters
            Fit parameters to be initialized.
        """
        sn_data = self.training_dataset.sn_data
        pars['X0'].full[sn_data.sn_index] = sn_data.x0
        pars['X1'].full[sn_data.sn_index] = sn_data.x1
        pars['c'].full[sn_data.sn_index] = sn_data.col
        pars['tmax'].full[sn_data.sn_index] = sn_data.tmax

    def init_model_pars(self, pars, model_name='salt2', version='2.4'):
        """Load and adapt the SALT2/SALT3 global surfaces and color law from `sncosmo`.

        This method loads the coefficients of the SALT2.4 surfaces and color
        law. The definition of the model spline bases may differ from the
        original definition of the SALT2/SALT3 bases; thus, the original
        surfaces are reprojected onto the model basis.

        Parameters
        ----------
        pars : nacl.lib.fitparameters.FitParameters
            Fit parameters to be initialized.
        model_name : str, optional
            The name of the model version to use for initialization, by default 'salt2'.
        version : str, optional
            The version of the model to use for initialization, by default '2.4'.
        """
        salt2_source = sncosmo.get_source('salt2', version='2.4')

        phase_grid = np.linspace(self.phase_range[0], self.phase_range[1], self.n_ph)
        wl_grid = np.linspace(self.wl_range[0], self.wl_range[1], self.n_wl)
        basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)

        phase_salt = salt2_source._phase
        wl_salt = salt2_source._wave

        w, p = np.meshgrid(wl_salt, phase_salt)

        sncosmo_scale = salt2_source._SCALE_FACTOR
        salt2_m0 = salt2_source._model['M0'](phase_salt, wl_salt).ravel() / sncosmo_scale
        salt2_m1 = salt2_source._model['M1'](phase_salt, wl_salt).ravel() / sncosmo_scale
        salt2_cl = np.array(salt2_source._colorlaw_coeffs)[::-1]

        jac = self.basis.eval(w.ravel(), p.ravel()).tocsr()
        factor = cholesky_AAt(jac.T, beta=1.E-6)

        pars['M0'].full[:] = factor(jac.T * salt2_m0)
        pars['M1'].full[:] = factor(jac.T * salt2_m1)
        pars['CL'].full[:] = salt2_cl

    @classmethod
    def _init_bases(cls, wl_grid=None, phase_grid=None,
                    wl_range=(2000., 9000.),
                    phase_range=(-20., 50.), basis_knots=(127,20)):
        """
        Instantiate model bases.

        This class method initializes the model bases using B-splines. If no
        wavelength or phase grid is provided, it creates default grids based
        on the specified ranges and number of knots.

        Parameters
        ----------
        wl_grid : array-like, optional
            User-provided grid in wavelength. If None, a default grid is created.
        phase_grid : array-like, optional
            User-provided grid in phase. If None, a default grid is created.
        wl_range : tuple, optional
            Wavelength range for the default grid, by default (2000., 9000.).
        phase_range : tuple, optional
            Phase range for the default grid, by default (-20., 50.).
        basis_knots : tuple, optional
            Number of knots in wavelength and phase, by default (127, 20).

        Returns
        -------
        bspline.BSpline2D
            The B-spline basis for the model.
        """
        if wl_grid is None:
            assert wl_range is not None and basis_knots is not None
            n_wl, n_ph = basis_knots
            logging.info('default regular grid in wavelength')
            wl_grid = np.linspace(wl_range[0], wl_range[1], n_wl)
        else:
            logging.info('user provided grid in wavelength')
        if phase_grid is None:
            assert phase_range is not None and basis_knots is not None
            n_wl, n_ph = basis_knots
            logging.info('default regular grid in phase')
            phase_grid = np.hstack([np.linspace(phase_range[0],
                                                phase_range[1],
                                                n_ph)])
        else:
            logging.info('user provided grid in phase')

        return bspline.BSpline2D(wl_grid, phase_grid,
                                 x_order=4, y_order=4)

    def normalization(self, pars=None, band_name='swope2::b', Mb=-19.5, magsys='AB',
                      default_norm=1.01907246e-12):
        """
        Model normalization.

        The SALT2Like normalization is set during training by the constraint
        on the integral of M0 at phase zero. By default, the model is not
        renormalized during evaluation.

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters, optional
            Model parameters for normalization. If None, the default normalization is returned.
        band_name : str, optional
            The band used for normalization, by default 'swope2::b'.
        Mb : float, optional
            Absolute magnitude, by default -19.5.
        magsys : str, optional
            Magnitude system, by default 'AB'.
        default_norm : float, optional
            Default normalization value, by default 1.01907246e-12.

        Returns
        -------
        float
            The normalization factor.
        """
        default_norm = default_norm if default_norm is not None else 1.01907246e-12
        # tq, filter_basis = self.filter_db[band_name]
        passband_data = self.filter_db[band_name]

        # AB flux in the specified band
        ms = SNMagSys(self.filter_db, magsys)
        # zp = ms.ZeroPoint(self.filter_db.transmission_db[band_name])
        zp = ms.get_zp(band_name)
        self.int_ab_spec = 10**(0.4 * zp)

        # normalization quantities
        self.flux_at_10pc = np.power(10., -0.4 * (Mb-zp))
        self.flux_at_10Mpc = np.power(10., -0.4 * (Mb+30.-zp))

        if pars is None:
            return default_norm

        phase_eval = self.basis.by.eval(np.array([0. + self.delta_phase])).tocsr()
        gram = get_gram(z=0., model_basis=self.basis.bx,
                        filter_basis=passband_data.basis, lambda_power=1)
        surface_0 = pars['M0'].full.reshape(len(self.basis.by), -1)
        # evaluate the integral of the model in the specified band
        self.int_M0_phase_0 = phase_eval.dot(surface_0.dot(gram.dot(passband_data.tq)))

        return self.flux_at_10Mpc / self.int_M0_phase_0

    def renorm(self, pars, band_name='swope2::b', Mb=-19.5, magsys='AB'):
        """
        Adjust the model normalization.

        This method explicitly recomputes the normalization factor based on the
        provided parameters, band, magnitude, and magnitude system.

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters
            Model parameters for normalization.
        band_name : str, optional
            The band used for normalization, by default 'swope2::b'.
        Mb : float, optional
            Absolute magnitude, by default -19.5.
        magsys : str, optional
            Magnitude system, by default 'AB'.

        Returns
        -------
        None
        """
        # explicitely recompute a normalization
        self.norm = self.normalization(pars, band_name, Mb, magsys,
                                       default_norm=None)

    def _init_eval_units(self):
        """
        Initialize the photometric and spectroscopic evaluation units.

        The actual evaluation of the model is delegated to specialized
        sub-units, which are called sequentially. These units are stored
        in a queue.

        Returns
        -------
        list
            List of initialized evaluation units.
        """
        queue = []
        if self.training_dataset.lc_data is not None:
            logging.info('initializing lightcurve eval unit')
            queue.append(LightcurveEvalUnit(self))
        if self.training_dataset.spec_data is not None:
            logging.info('initializing spectrum eval unit')
            queue.append(CompressedSpectrumEvalUnit(self, spec_recal_degree=self.spectrum_recal_degree))
        if self.training_dataset.spectrophotometric_data is not None:
            logging.info('initializing spectrophotometric eval unit')
            queue.append(CompressedSpectroPhotoEvalUnit(self))
        return queue

    def precompute_color_law(self, cl_pars, jac=False):
        """
        Precompute the color law for the photometric data.

        Parameters
        ----------
        cl_pars : numpy.array or None
            Color law parameters.
        jac : bool, optional
            If True, computes and stores the Jacobian of the color law, by default False.

        Returns
        -------
        None
        """
        polynome_color_law, jacobian_color_law = \
            self.color_law(self.L_eff.data, cl_pars,
                           jac=jac)  # return_jacobian_as_coo_matrix=False)
        self.polynome_color_law = \
            scipy.sparse.csr_matrix((polynome_color_law,
                                     (self.L_eff.row, self.L_eff.col)),
                                    shape=self.L_eff.shape)
        if jac:
            self.jacobian_color_law = []
            _, n = jacobian_color_law.shape
            for i in range(n):
                jacobian_cl = \
                    scipy.sparse.csr_matrix((jacobian_color_law[:, i],
                                             (self.L_eff.row, self.L_eff.col)), shape=self.L_eff.shape)
                self.jacobian_color_law.append(jacobian_cl)

    def get_restframe_phases(self, pars, data):
        """
        Compute the rest-frame phases for a given dataset.

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters
            Model parameters containing 'tmax'.
        data : object
            Dataset containing MJD and redshift information.

        Returns
        -------
        numpy.array
            Supernova rest-frame phases.
        """
        sn_index = data.sn_index
        tmax = pars['tmax'].full[sn_index]
        return (data.mjd - tmax) / (1.+data.z)

    def __call__(self, pars, jac=False): # , plotting=False, ilc_plot=None, spec_plot=False):
        """
        Evaluate the model for the given parameter set.

        This method evaluates the model by looping over the evaluation units
        and assembling the results into a single vector and, optionally, a
        single Jacobian matrix.

        Parameters
        ----------
        pars : numpy.array
            Vector containing the free parameters.
        jac : bool, optional
            If True, returns the Jacobian matrix, by default False.

        Returns
        -------
        numpy.array
            Model evaluation results.
        scipy.sparse.csr_matrix, optional
            Jacobian matrix if `jac` is True.
        """
        cached_res = self._get_from_cache(pars, jac=jac)
        if cached_res is not None:
            logging.debug('SALT2Like.__call__: returning cached result')
            self.cached = True
            return cached_res
        else:
            self.cached = False

        # self.clear_cache()
        # loop over the eval units
        res = [q(pars, jac) for q in self.queue]

        if not jac:
            model_val = np.add.reduce([r for r in res])
            self._save_to_cache(pars, model_val, None)
            return model_val

        model_val = np.add.reduce([r[0] for r in res])

        rows = np.hstack([r[1].row for r in res])
        cols = np.hstack([r[1].col for r in res])
        vals = np.hstack([r[1].data for r in res])

        idx = cols >= 0
        JJ = scipy.sparse.coo_matrix((vals[idx], (rows[idx], cols[idx])),
                                     shape=res[0][1].shape)

        self._save_to_cache(pars, model_val, JJ)
        return model_val, JJ

    @property
    def y(self):
        """
        Get all flux measurements from the training dataset.

        Returns
        -------
        numpy.array
            Array of all flux measurements.
        """
        return self.training_dataset.get_all_fluxes()

    @property
    def yerr(self):
        """
        Get all flux measurement errors from the training dataset.

        Returns
        -------
        numpy.array
            Array of all flux measurement errors.
        """
        return self.training_dataset.get_all_fluxerr()

    @property
    def bads(self):
        """
        Get boolean array flagging the bad data points from the training dataset.

        Returns
        -------
        numpy.array
            Array of indices of bad data points.
        """
        return self.training_dataset.get_bads()


_G, _G2 = {}, {}


def get_gram(z, model_basis, filter_basis, lambda_power=1):
    """
    Calculate the grammian of to spline basis.

    The grammian :math:`G` of order :math:`N`, to basis :math:`B_0` (define on wavelength,
    in our case surfaces wavelength basis) and
    :math:`B_1` (define on SN-restframe wavelength, filter basis) is defined as :

    .. math::
        G = \\int \\lambda^N B_0(\\lambda) B_q(\\lambda (1+z)) d\\lambda

    Parameters
    -------
    z : numpy.array
        Vector containing the data redshift.
    model_basis : nacl.lib.bspline.BSpline
        Wavelength basis.
    filter_basis : nacl.lib.bspline.BSpline
        Filter basis defined common to all the bands.
    lambda_power : int
        Gramian order.

    Returns
    -------
    gram : scipy.sparse.csc_matrix
        Grammian
    """
    global _G
    global _G2
    key = (z, model_basis, filter_basis, lambda_power)

    gram = _G.get(key, None)
    if gram is None:
        gram = bspline.lgram(model_basis, filter_basis, z=z, lambda_power=lambda_power).tocsc()
        _G[key] = gram

    return gram



"""NaCl SALT2Like error models
"""

import logging

import numpy as np
import numpy.polynomial.polynomial as nppol
from scipy import sparse

from bbf.bspline import BSpline2D
from nacl.sparseutils import CooMatrixBuff2

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt


class SimpleErrorSnake:
    r"""
    A simplistic error snake model with a single parameter.

    This model represents the variance as a function of the model values:

    .. math::
        V = g^2 \cdot \text{model}^2

    where `g` is a parameter representing the scaling factor.

    Parameters
    ----------
    model : object
        The model used for evaluation.
    """
    def __init__(self, model):
        """
        Constructor.

        Parameters
        ----------
        model : object
            The model used for evaluation.
        """
        self.model = model

    def get_struct(self):
        """
        Get the block structure of the part of the parameter vector used to evaluate the model

        Returns
        -------
        list of tuple
            List containing the block structure of the parameter vector.
            Each item contains the block name and size.
        """
        return [("gamma", 1)]

    def init_pars(self, pars):
        """
        Initialize the blocks of the parameter vector which are specific to this class.

        Parameters
        ----------
        pars : `nacl.fitparameters.FitParameters`
            Parameters object to be initialized.
        """
        pars['gamma'].full[:] = 0.01

    def __call__(self, pars, jac=False):
        """
        Evaluate the variance model and optionally its Jacobian.

        Parameters
        ----------
        pars : object
            Parameters object containing the model parameters.
        jac : bool, optional
            If True, compute and return the Jacobian matrix. Default is False.

        Returns
        -------
        var : numpy.ndarray
            Array of variances.
        J : scipy.sparse.coo_matrix, optional
            Jacobian matrix, returned if `jac` is True.
        """
        v, J = None, None
        if jac:
            v, J = self.model(pars, jac=True)
        else:
            v = self.model(pars, jac=False)
        N = len(v)
        g = pars['gamma'].full[0]
        var = np.full(N, g**2 * v**2)

        if not jac:
            return var

        J = J.tocoo()

        n_free_pars = len(pars.free)
        i = [J.row]
        j = [J.col]
        data = [2. * g**2 * v[J.row] * J.data]
        i.append(np.arange(N))
        j.append(np.full(N, pars['gamma'].indexof(np.zeros(N).astype(int))))
        data.append(2. * g * v**2)

        i = np.hstack(i)
        j = np.hstack(j)
        data = np.hstack(data)

        idx = j >= 0
        J = sparse.coo_matrix((data[idx], (i[idx], j[idx])),
                              shape=(N, n_free_pars))

        return var, J


class LocalErrorSnake:
    r"""
    Error snake variance as a function of phase and wavelength

    This model represents the variance as a combination of a fixed error term and a model-dependent error term:

    .. math::
        V(\lambda, p) = \sigma_{Err}^2 + \sigma_{\mathrm{Mod}}^2(\lambda, p)

    where:

    .. math::
        \sigma_{\mathrm{Mod}}^{spec} &= g(\lambda, p) \times \phi_{\mathrm{spec}}(p, \lambda)\ \ (\mathrm{spectra}) \\
        \sigma_{\mathrm{Mod}}^{phot} &= g(\lambda, p) \times \phi_{\mathrm{phot}}(p, \lambda)\ \ (\mathrm{photometric})

    :math:`\gamma(\lambda, p)` is a global surface describing the spectral residual of a 2D spline surface defined on the same ranges as the corresponding model.
    """
    def __init__(self, model, bins=(10, 10), order=4):
        """Constructor

        Parameters
        ----------
        model : object
            The model used for evaluation.
        bins : tuple of int, optional
            The number of bins for the phase and wavelength grids (default is (10, 10)).
        order : int, optional
            The order of the B-spline (default is 4).
        """
        self.model = model
        self.training_dataset = model.training_dataset

        # instantiate a basis
        ph_bins, wl_bins = bins
        phase_grid = np.linspace(model.basis.by.range[0], model.basis.by.range[-1], ph_bins)
        wl_grid = np.linspace(model.basis.bx.range[0], model.basis.bx.range[-1], wl_bins)
        self.basis = BSpline2D(wl_grid, phase_grid, x_order=order, y_order=order)

    def get_struct(self):
        """
        Get the block structure of the part of the parameter vector used to evaluate the model.

        Returns
        -------
        list of tuple
            List containing the block structure of the parameter vector.
            Each item contains the block name and size.
        """
        return [('gamma', len(self.basis))]

    def init_pars(self, pars):
        """
        Initialize the blocks of the parameter vector which are specific to this class.

        Parameters
        ----------
        pars : `nacl.fitparameters.FitParameters`
            Parameters object to be (partially) initialized.
        """
        pars['gamma'].full[:] = 0.01

    def __call__(self, pars, jac=False):
        """
        Evaluate the variance model and optionally its Jacobian.

        Parameters
        ----------
        pars : object
            Parameters object containing the model parameters.
        jac : bool, optional
            If True, compute and return the Jacobian matrix. Default is False.

        Returns
        -------
        var : numpy.ndarray
            Array of variances.
        J : scipy.sparse.coo_matrix, optional
            Jacobian matrix, returned if `jac` is True.
        """
        # too bad we need to call the model again here...
        # let's see whether we could cache this call
        if jac:
            flux, Jm = self.model(pars, jac=True)
        else:
            flux = self.model(pars, jac=False)
            Jm = None

        tds = self.training_dataset
        zz = tds.lc_data.z
        tmax = pars['tmax'].full[tds.lc_data.sn_index]
        wl = tds.lc_data.wavelength / (1. + zz)
        phase = (tds.lc_data.mjd - tmax) / (1. + zz)
        Jv = self.basis.eval(wl, phase)

        g = np.zeros(len(tds))
        g[:len(tds.lc_data)] = np.exp(Jv @ pars['gamma'].full)
        v = g * flux**2
        if not jac:
            return v

        # derivatives
        N = len(tds)
        n_free_pars = len(pars.free)
        buff = CooMatrixBuff2((N, n_free_pars))

        # d_gamma
        buff.append(Jv.row,
                    pars['gamma'].indexof(Jv.col),
                    g[Jv.row] * Jv.data * flux[Jv.row]**2)

        # d_dpars
        buff.append(Jm.row,
                    Jm.col,
                    2. * g[Jm.row] * flux[Jm.row] * Jm.data)

        J = buff.tocoo()

        return v, J



class SNLambdaErrorSnake:
    """
    A model to evaluate the variance as a function of wavelength for supernovae.

    This model includes a polynomial dependency on the wavelength for the error calculation.
    """
    def __init__(self, model, deg=3):
        """
        Constructor.

        Parameters
        ----------
        model : object
            The model used for evaluation.
        deg : int, optional
            The degree of the polynomial used for the wavelength dependency (default is 3).
        """
        self.model = model
        self.training_dataset = self.model.training_dataset
        self.deg = deg

        #pre construction of the vandermonde matrix
        lc_data = self.training_dataset.lc_data
        restframe_wl = lc_data.wavelength / (1. + lc_data.z)
        r_wl = np.zeros(len(self.training_dataset))
        r_wl[:len(lc_data)] = restframe_wl
        self.vander_matrix = self.get_vander_matrix(r_wl)

    def get_struct(self):
        """
        Get the block structure of the part of the parameter vector relevant for model evaluation.

        Returns
        -------
        list of tuple
            List containing the block structure of the parameter vector.
            Each item contains the block name and size.
        """
        nsn = len(self.training_dataset.sn_data.sn_set)
        npars = self.deg # deg + 1 - 1 constraint
        return [('gamma_sn', nsn), ('gamma_snake', npars)]

    def init_pars(self, pars):
        """
        Initialize the blocks of the parameter vector which are specific to this class.

        Parameters
        ----------
        pars : `nacl.fitparameters.FitParameters`
            Parameters object to be (partially) initialized.
        """
        pars['gamma_sn'].full[:] = 0.01
        pars['gamma_snake'].full[:] = 0.

    def _reduce(self, restframe_wl):
        """
        Map the restframe wavelength to a standardized range (close to [-1,1])

        This is done mostly to improve the conditioning of the polynomial fit.

        Parameters
        ----------
        restframe_wl : numpy.ndarray
            Array of restframe wavelengths.

        Returns
        -------
        numpy.ndarray
            Reduced wavelengths.
        """
        return (restframe_wl-4500.) / (9000. - 2000.)

    def get_vander_matrix(self, restframe_wl):
        """
        Get the Vandermonde matrix for the reduced wavelengths.

        Parameters
        ----------
        restframe_wl : numpy.ndarray
            Array of restframe wavelengths.

        Returns
        -------
        numpy.ndarray
            Vandermonde matrix for the polynomial fit.
        """
        reduced_wl = self._reduce(restframe_wl)
        vander_mat = np.vander(reduced_wl, self.deg, increasing = False)
        return vander_mat

    def __call__(self, pars, jac=False):
        """
        Evaluate the variance model and optionally its Jacobian.

        Parameters
        ----------
        pars : object
            Parameters object containing the model parameters.
        jac : bool, optional
            If True, compute and return the Jacobian matrix. Default is False.

        Returns
        -------
        variance : numpy.ndarray
            Array of variances.
        J : scipy.sparse.coo_matrix, optional
            Jacobian matrix, returned if `jac` is True.
        """
        if jac:
            flux, Jm = self.model(pars, jac=True)
        else:
            flux = self.model(pars, jac=False)
            Jm = None

        tds = self.training_dataset
        zz = tds.lc_data.z
        restframe_wl = tds.lc_data.wavelength / (1. + zz)
        reduced_wl = self._reduce(restframe_wl)
        snake = np.zeros(len(tds))
        snake[:len(tds.lc_data)] = np.exp(np.polyval(pars['gamma_snake'].full, reduced_wl))

        g = np.zeros(len(tds))
        g[:len(tds.lc_data)] = pars['gamma_sn'].full[tds.lc_data.sn_index]
        variance = g**2 * snake * flux**2

        if not jac:
            return variance

        # derivatives
        N = len(tds)
        n_free_pars = len(pars.free)
        buff = CooMatrixBuff2((N, n_free_pars))

        # d_gamma = 2 * snake * g_i * flux**2
        buff.append(tds.lc_data.row,
                    pars['gamma_sn'].indexof(tds.lc_data.sn_index),
                    2 * snake[tds.lc_data.row] * g[tds.lc_data.row] * flux[tds.lc_data.row]**2)

        # d_snake = g**2 * flux**2 * reduced_wl ** i * snake
        data_d_snake = sparse.coo_matrix(variance[:,np.newaxis] * self.vander_matrix[:,0:self.deg-1])
        row_d_snake = data_d_snake.row
        col_d_snake = pars['gamma_snake'].indexof(data_d_snake.col)
        buff.append(row_d_snake,
                    col_d_snake,
                    data_d_snake.data)

        # d_pars = 2 * snake * g**2 + flux * Jm
        buff.append(Jm.row,
                    Jm.col,
                    2. * g[Jm.row]**2 * snake[Jm.row] * flux[Jm.row] * Jm.data)

        J = buff.tocoo()

        return variance, J


class SNLocalErrorSnake:
    r"""
    Error snake variance model as a function of phase and wavelength.

    This model evaluates the variance using a combination of phase, wavelength,
    and a de-weighting factor for each supernova (SN) to capture SN variability.

    .. math::
        V(\lambda, p) = \sigma_{Err}^2 + \sigma_{\mathrm{Mod}}^2(\lambda,p) \times \sigma_{\mathrm{SN}}

    where:

    .. math::
        \sigma_{\mathrm{Mod}}^{spec} &= g(\lambda, p) \times \gamma_{\mathrm{SN}} \times \phi_{\mathrm{spec}}(p,\lambda) \ (\mathrm{spectra}) \\
        \sigma_{\mathrm{Mod}}^{phot} &= g(\lambda, p) \times \gamma_{\mathrm{SN}} \times \phi_{\mathrm{phot}}(p,\lambda) \ (\mathrm{photometric})

    :math:`\gamma(\lambda, p)` is a global surface describing the spectral residual
    of a 2D spline surface defined on the same ranges as the corresponding model.
    :math:`\gamma_{\mathrm{SN}}` is a de-weighting factor for each SN to capture SN variability.
    """
    def __init__(self, model, bins=(10, 10), order=4):
        """
        Constructor for the SNLocalErrorSnake class.

        Parameters
        ----------
        model : object
            The model used for evaluation.
        bins : tuple of int, optional
            Number of bins for phase and wavelength. Default is (10, 10).
        order : int, optional
            Order of the B-spline. Default is 4.
        """
        self.model = model
        self.training_dataset = model.training_dataset

        # instantiate a basis
        ph_bins, wl_bins = bins
        phase_grid = np.linspace(model.basis.by.range[0], model.basis.by.range[-1], ph_bins)
        wl_grid = np.linspace(model.basis.bx.range[0], model.basis.bx.range[-1], wl_bins)
        self.basis = BSpline2D(wl_grid, phase_grid, x_order=order, y_order=order)

    def get_struct(self):
        """
        Get the block structure of the parameter vector used to evaluate the model.

        Returns
        -------
        list of tuple
            List containing the block structure of the parameter vector.
            Each item contains the block name and size.
        """
        nsn = len(self.training_dataset.sn_data.sn_set)
        return [('gamma_snake', len(self.basis)), ('gamma_sn', nsn)]

    def init_pars(self, pars):
        """
        Initialize the blocks of the parameter vector which are specific to this class.

        Parameters
        ----------
        pars : `nacl.fitparameters.FitParameters`
            Parameters object to be (partially) initialized.
        """
        pars['gamma_sn'].full[:] = 0.01
        pars['gamma_snake'].full[:] = 0.01

    def __call__(self, pars, jac=False):
        """
        Evaluate the variance model and optionally its derivatives.

        Parameters
        ----------
        pars : object
            Parameters object containing the model parameters.
        jac : bool, optional
            If True, compute and return the Jacobian matrix. Default is False.

        Returns
        -------
        v : numpy.ndarray
            Array of variances.
        J : scipy.sparse.coo_matrix, optional
            Jacobian matrix, returned if `jac` is True.
        """
        if jac:
            flux, Jm = self.model(pars, jac=True)
        else:
            flux = self.model(pars, jac=False)
            Jm = None

        tds = self.training_dataset
        zz = tds.lc_data.z
        tmax = pars['tmax'].full[tds.lc_data.sn_index]
        wl = tds.lc_data.wavelength / (1. + zz)
        phase = (tds.lc_data.mjd - tmax) / (1. + zz)
        Jv = self.basis.eval(wl, phase)

        # fixing phase edges parameters
        njx = self.basis.bx.nj
        nj = self.basis.nj
        cols_to_keep = np.arange(0, nj-njx-1, 1)

        #cols_full = np.arange(0, nj, 1)
        #b = np.array([0, njy -1])
        #c = np.arange(0, njy, 1)
        #a = c*njy
        #x = [b+i for i in a]
        #x = np.hstack(x)
        #cols_remove = np.isin(cols_full, x)
        #cols_to_keep = ~cols_remove

        mask = np.isin(Jv.col, cols_to_keep)

        filtered_data = Jv.data[mask]
        filtered_row = Jv.row[mask]
        filtered_col = Jv.col[mask]

        col_map = {old_col:new_col for new_col, old_col in enumerate(cols_to_keep)}
        filtered_col = np.array([ col_map[c] for c in filtered_col ])
        new_shape = (Jv.shape[0], len(cols_to_keep))
        Jv = sparse.coo_matrix( (filtered_data, (filtered_row, filtered_col)), shape=new_shape)

        #
        g_snake = np.zeros(len(tds))
        g_snake[:len(tds.lc_data)] = np.exp(Jv @ pars['gamma_snake'].full[:-njx-1])

        g_sn = np.zeros(len(tds))
        g_sn[:len(tds.lc_data)] = pars['gamma_sn'].full[tds.lc_data.sn_index]

        v = g_snake * g_sn**2 * flux**2
        if not jac:
            return v

        # derivatives
        N = len(tds)
        n_free_pars = len(pars.free)
        buff = CooMatrixBuff2((N, n_free_pars))

        # d_gamma_snake
        buff.append(Jv.row,
                    pars['gamma_snake'].indexof(Jv.col),
                    g_snake[Jv.row] * Jv.data * g_sn[Jv.row]**2 * flux[Jv.row]**2)

        # d_gamma_sn
        buff.append(tds.lc_data.row,
                    pars['gamma_sn'].indexof(tds.lc_data.sn_index),
                    2 * g_sn[tds.lc_data.row] * g_snake[tds.lc_data.row] * flux[tds.lc_data.row]**2)

        # d_dpars
        buff.append(Jm.row,
                    Jm.col,
                    2. * g_snake[Jm.row] * g_sn[Jm.row]**2 * flux[Jm.row] * Jm.data)

        J = buff.tocoo()

        return v, J


class SimpleErrorSnake_:
    r"""Simple, 1 parameter error snake: :math:`\sigma_f = \sigma_m \times f`

    This is the simplest error snake possible. We add a pedestal variance,
    constant in magnitude (i.e. of the form :math:`\sigma_m \times flux` in
    flux).
    """

    def __init__(self, model):  # , var_pedestal=0.001):
        self.model = model
        self.training_dataset = self.model.training_dataset
        # self.var_pedestal = var_pedestal

    def get_struct(self):
        """
        """
        return [('sigma_snake', 1)]

    def __call__(self, pars, jac=False, model_flux=None, model_jac=None):
        """return the diagonal of the error snake variance + derivatives
        """
        mflux, mjac = None, None
        # if parameters are given, we need to re-evaluate the model
        if p is not None:
            if not jac:
                mflux = self.model(p)
            else:
                mflux, mjac = self.model(p, jac=True)
        # otherwise, we re-evaluate it only if it is necessary
        else:
            if not jac:
                mflux = model_flux if model_flux is not None else self.model(self.model.pars.free)
            else:
                if model_flux is not None and model_jac is not None:
                    mflux, mjac = model_flux, model_jac
                else:
                    mflux, mjac = self.model(self.model.pars.free, jac=True)

        # model parameters
        pars = self.model.pars
        sigma_m = self.model.pars['sigma_snake'].full[0]

        # evaluate the variance
        var = (sigma_m * mflux)**2

        # we should only apply it to the photometric measurements
        # nlc, _, _ = self.model.training_dataset.n

        # and return it if this is the only thing requested
        if not jac:
            return var

        # derivatives
        n = len(self.model.pars.full)
        n_free = len(self.model.pars.free)
        N = self.training_dataset.nb_meas(valid_only=False)

        # dV / dsigma_snake
        i = np.arange(N).astype(int)
        j = np.full(N, pars['sigma_snake'].indexof())
        dVds = 2. * sigma_m * mflux**2
        idx = j>=0
        J_dVds = sparse.coo_matrix((dVds[idx], (i[idx],j[idx])), shape=(N,n_free))

        # dV/dbeta
        # mjac.data *= (2. * sigma_m**2 * mflux[mjac.row])
        J_dVdbeta = sparse.coo_matrix((mjac.data * 2. * sigma_m**2 * mflux[mjac.row],
                                (mjac.row, mjac.col)),
                                shape=mjac.shape)

        # i = np.hstack((i, mjac.row))
        # j = np.hstack((j, mjac.col))
        # v = np.hstack((dVds, mjac.data))
        # idx = j>=0
        # J = coo_matrix((v[idx], (i[idx],j[idx])), shape=(N, n_free))
        # J = J_dVds.tocsc() + J_dVdbeta.tocsc()
        J = J_dVds.tocoo() + J_dVdbeta.tocoo()

        return var, J

    def noise(self, p=None, **kwargs):
        """draw one realization of the error snake noise
        """
        var = self(p, jac=False, **kwargs)
        nlc, nsp, nphotspec = self.training_dataset.nb_meas(valid_only=False,
                                                            split_by_type=True)
        N = nlc + nsp + nphotspec
        n = np.zeros(N)
        # this error snake only applies to the photometric dataset
        n[:nlc] = np.random.normal(loc=0., scale=np.sqrt(var[:nlc]))
        return n


class CalibErrorModel:

    def __init__(self, model, calib_variance,
                 default_calib_variance=0.5**2):
        """Constructor

        Parameters
        ----------
        model : a NaCl model (e.g. SALT2Like)
            the underlying model
        calib_variance : float or pandas.DataFrame
            the calibration covariance matrix. It is either a float, if all
            bands have the same calibration variance (uncorrelated), or if all
            calibration uncertainties are uncorrelated, the diagonal of the
            covariance matrix, or the full covariance matrix.
        """
        self.model = model
        self.training_dataset = self.model.training_dataset
        self.calib_variance = calib_variance
        self.default_calib_variance = default_calib_variance
        self.inv_full_covmat = None
        self.sub_covmat = None

    def get_struct(self):
        nb_bands = self.training_dataset.nb_bands()
        return [('eta_calib', nb_bands)]
    def init_pars(self, pars):
        """
        """
        pars['eta_calib'].full[:] = 1.

    def build_covmat(self, pars,
                     calib_variance, bands=None,
                     default_calib_variance=0.5):
        """build a covmat (ordered to match the lc_data.band_index)

        Returns
        -------
        np.ndarray
            the re-ordered covmat
        """
        npars = len(pars.full)
        pfull = pars.copy()
        pfull.release()

        sz = len(pfull['eta_calib'].full)

        # if all bands have the same variance, that's easy
        if isinstance(calib_variance, float):
            nb_bands = self.training_dataset.nb_bands()
            i = pfull['eta_calib'].indexof()
            v = np.full(sz, 1./self.calib_variance)
            self.sub_covmat = np.diag(np.full(nb_bands, calib_variance))
            self.inv_full_covmat = sparse.coo_matrix((v, (i, i)), shape=(npars, npars)).tocsr()

        # otherwise, we may get a dict of the form {<band_name>: var, ...}
        # in which case it is also easy
        elif isinstance(calib_variance, dict):
            tds = self.model.training_dataset
            diag = []
            for bn in tds.lc_data.band_set:
                if bn not in calib_variance:
                    logging.warning(f'no calibration variance specified for band: {bn} -- will assign {default_calib_variance} by default')
                diag.append(calib_variance.get(bn, default_calib_variance))
            diag = np.array(diag)
            assert np.all(diag)
            i = pfull['eta_calib'].indexof()
            self.sub_covmat = np.diag(diag)
            self.inv_full_covmat = sparse.coo_matrix((1./diag, (i,i)), shape=(npars, npars)).tocsr()

        # finally, we may get a real matrix, along with a dict that
        # specifies the indices to band mapping. Then, all we have
        # to do, is to make sure all the bands are here, fill out the blank
        # if necessary, and re-order it if necessary
        elif isinstance(calib_variance, tuple):
            covmat, bands  = calib_variance
            # extend band into a simple map
            band_map = dict([(y,x) for x,y in enumerate(bands)])

            # extend calib covmat with a defaut value
            nr,nc = covmat.shape
            assert nr == nc
            extended_covmat = np.zeros((nr+1,nc+1))
            extended_covmat[:nr,:nc] = covmat
            extended_covmat[nr,nr] = default_calib_variance
            #            print('extended_covmat: ', extended_covmat)
            # re-order covmat
            i = []
            for bn in self.model.training_dataset.lc_data.band_set:
                if bn not in bands:
                    logging.warning('no calibration variance specified for band: {bn} -- will assign {default_calib_variance} by default')
                    i.append(nr)
                else:
                    i.append(band_map[bn])
            ii,jj = np.meshgrid(i,i)
            reordered_covmat = extended_covmat[ii,jj]

            # invert the covmat
            w = np.linalg.inv(reordered_covmat)

            # and re-index it into a larger (n,n) matrix, where n = number of parameters
            nn, _ = w.shape
            ii, jj = np.mgrid[:nn,:nn]
            self.sub_covmat = reordered_covmat
            self.inv_full_covmat = sparse.coo_matrix((w.flatten(),
                                               (pfull['eta_calib'].indexof(ii.flatten()),
                                                pfull['eta_calib'].indexof(jj.flatten()))),
                                              shape=(npars, npars)).tocsr()
        else:
            raise ValueError(f"build_covmat(): don't know what to do with {calib_variance}")

    def __call__(self, pars, deriv=False):
        """evaluate the calib prior (and optionally its gradient and Hessian)
        """
        #        if p is not None:
        #            self.model.pars.free = p

        if self.sub_covmat is None or self.inv_full_covmat is None or self.inv_full_covmat.shape[0]!=pars.full.shape[0]:
            self.build_covmat(pars, self.calib_variance)

        pfull = pars.full
        prior = np.dot(pfull, (self.inv_full_covmat.dot(pfull)))
        if not deriv:
            return prior

        # TODO: add some sanity and change the sign in the LogLikelihood
        # so that can return grad and not -grad ...
        # TODO: I am here ! -> reduce size, depending on fixed parameters

        n = self.inv_full_covmat.shape[0]
        idx = pars.indexof(np.arange(n)) >= 0

        grad = 2. * (self.inv_full_covmat @ pfull)[idx]
        hess = 2. * self.inv_full_covmat[:,idx][idx,:]

        return prior, grad, hess

    def correction(self, pars, lcdata=None):
        """the multiplicative correction to apply to the photometric data

        Parameters
        ----------
        p: (ndarray), optional
          vector of free parameters
        lcdata: stff, optional
          something that has a band_index

        Returns:
          a ndarray with the corrections
        """
#        if p is not None:
#            self.model.pars.free = p
        if lcdata is None:
            lcdata = self.training_dataset.lc_data
        corr = 1. + pars['eta_calib'].full[lcdata.band_index]
        return corr

    def noise(self, pars):
        """the noise component, to apply to the photometric data
        """
        # if p is not None:
        #     self.model.pars.free = p

        lc_data = self.training_dataset.lc_data

        nphot, nsp, nspphot = self.training_dataset.nb_meas(valid_only=False,
                                                            split_by_type=True)
        scales = np.ones(nphot + nsp + nspphot)
        nb_bands = self.training_dataset.nb_bands()

        # covmat = self.build_covmat()
        L = np.linalg.cholesky(self.sub_covmat)
        dm = L @ np.random.normal(0., scale=1., size=nb_bands)
        # dm = [dm[self.band_map[i]] for i in range(len(self.bands))]
        scales[:nphot] = 1. + dm[lc_data.band_index]
        return scales # 1.+ dm[lc_data.band_index]


class ColorScatterModel:
    """An adaptation of the color scatter model developed by Guy Augarde
    """
    U_WAVELENGTH = 3650.88
    B_WAVELENGTH = 4302.57
    V_WAVELENGTH = 5428.55
    R_WAVELENGTH = 6418.01
    I_WAVELENGTH = 7968.34
    WAVELENGTH = {"U": U_WAVELENGTH, "B": B_WAVELENGTH, "V": V_WAVELENGTH,
                  "R": R_WAVELENGTH, "I": I_WAVELENGTH}

    def __init__(self, model):
        """Constructor
        """
        self.model = model
        self.training_dataset = model.training_dataset
        lc_data = self.training_dataset.lc_data
        restframe_wl = lc_data.wavelength / (1. + lc_data.z)
        self.reduced_restframe_wl = self.model.reduce(restframe_wl)

    def get_struct(self):
        """
        """
        nb_lightcurves = self.training_dataset.nb_lcs()
        return [('sigma_kappa', 3),
                ('kappa_color', nb_lightcurves)]

    def __call__(self, p=None, deriv=False):
        """return the color scatter prior, and (optionaly) its grad and hessian
        """
        pass

    def correction(self, pars):
        """return the multiplicative correction to be applied to the model
        """
        # if p is not None:
        #    self.model.pars.free = p
        tds = self.training_dataset
        corr = 1. + pars['kappa_color'].full[tds.lc_data.lc_index]
        return corr

    def noise(self, pars):
        """generate one realization of the noise -- to add directly to the data
        """
        # if p is not None:
        #     self.model.pars.free = p
        var = self.__call__(pars, jac=False)
        nn = np.random.normal(scale=np.sqrt(var), size=len(var))
        return 1. * nn[self.training_dataset.lc_data.lc_index]


class ColorScatter:
    r"""color scatter model

    To model the residual variability of the SNe color not described by the
    model, we allow the relative amplitude of each light curve to vary by a
    quantity :math:`(1+kappa)`:

    .. math::
        \phi_{phot}(p_0) = X_0 (1+z) (1+kappa) \int S(\lambda, p) T\left( \lambda \right) \frac{\lambda}{hc} d\lambda

    where :math:`\kappa` is an additional parameter, depending on the SN and the
    observation band.

    The variability of :math:`\kappa` is defined by a Gaussian prior, whose
    variance is a function of the wavelength restframe, and must be determined
    during the adjustment. In our model, this term is implemented by a
    polynomial of the wavelength restframe of the SN. For a light curve observed
    in a band :math:`X`, of average wavelength :math:`\lambda^X`, the
    corresponding variance is:

    .. math::
        \sigma_\kappa^2 = P\left(\frac{\lambda^X}{1+z}\right)

    where :math:`z` is redshift of the supernova and :math:`P` is a polynomial
    of the wavelength restframe. In practice, :math:`P` is implemented in terms
    of a reduced restframe wavelength: :math:`\lambda_{r}`

    Attributes
    ----------
    WL_REDUCED : numpy.array
        Reduced SN-restframe mean wavelength
    pars : nacl.lib.fitparameters.FitParameters
        Color scatter parameters
    """
    U_WAVELENGTH = 3650.88
    B_WAVELENGTH = 4302.57
    V_WAVELENGTH = 5428.55
    R_WAVELENGTH = 6418.01
    I_WAVELENGTH = 7968.34
    WAVELENGTH = {"U": U_WAVELENGTH, "B": B_WAVELENGTH, "V": V_WAVELENGTH,
                  "R": R_WAVELENGTH, "I": I_WAVELENGTH}

    def __init__(self, model): # wavelength_rest_reduced, sigma_kappa):
        """Constructor

        Parameters
        ----------
        wavelength_rest_reduced :
            Reduced SN-restframe mean wavelength
        sigma_kappa : numpy.array
            Color scatter parameter initialisation.
        """
        self.model = model
        self.training_dataset = model.training_dataset
        lc_data = self.training_dataset.lc_data
        restframe_wl = lc_data.wavelength / (1. + lc_data.z)
        _, ii = np.unique(lc_data.lc_index, return_index=True)
        ii.sort()
        self.reduced_restframe_wl = self.reduce(restframe_wl[ii])

    def reduce(self, wl):
        """
        """
        B_WL, V_WL = self.WAVELENGTH["B"], self.WAVELENGTH["V"]
        return (wl-B_WL)/(V_WL-B_WL)

    def get_struct(self):
        """
        """
        nb_lightcurves = self.training_dataset.nb_lcs()
        return [('sigma_kappa', 3),
                ('kappa_color', nb_lightcurves)]
    def init_pars(self, pars):
        """
        """
        pars['kappa_color'].full[:] = 1.E-6
        pars['sigma_kappa'].full[:] = 1.E-6
        pars['sigma_kappa'].full[0] = -4.

    def __call__(self, pars, deriv=False):
        r"""Return the color scatter variance (the :math:`V_\kappa` matrix)

        Parameters
        ----------
        p : np.ndarray
          free parameter vector

        Returns
        -------
        val : numpy.ndarray
            the diagonal of the color scatter variance matrix (used in the kappa prior)
        if jac:
            jacobian of the color scatter.
        """
        sigma_kappa = pars['sigma_kappa'].full
        kappa = pars['kappa_color'].full

        var = np.exp(nppol.Polynomial(sigma_kappa)(self.reduced_restframe_wl))

        # log det, penaly
        log_det = np.log(var).sum()
        quad_penalty = (kappa**2 / var).sum()
        penalty = log_det + quad_penalty

        if not deriv:
            return penalty

        # gradient w.r.t kappa
        N = len(pars.free)
        grad = np.zeros(N)
        j = pars['kappa_color'].indexof()
        idx = j>=0
        grad[j[idx]] = (2. * kappa / var)[idx]

        # gradient w.r.t sigma_kappa
        nk = len(pars['kappa_color'].full)
        ns = len(pars['sigma_kappa'].full)
        var = var.reshape(-1,1)
        kappa = kappa.reshape(-1,1)
        dq_ds = nppol.polyvander(self.reduced_restframe_wl, deg=2)
        tr_wdv =  np.sum(dq_ds, axis=0)
        kwdvwk = -np.sum(kappa**2 * dq_ds / var, axis=0)
        print(tr_wdv, kwdvwk)
        j = pars['sigma_kappa'].indexof()
        idx = j >= 0
        # check the signs
        grad[j[idx]] = -1*(-tr_wdv - kwdvwk)[idx]

        # hessian w.r.t kappa
        i, j, val = [], [], []
        m = sparse.coo_matrix(np.diag(2. / var.flatten()), shape=(nk,nk))
        i.append(pars['kappa_color'].indexof(m.row))
        j.append(pars['kappa_color'].indexof(m.col))
        val.append(m.data)

        # hessian: kappa-sigma_kappa block
        m = sparse.coo_matrix(-2. * kappa * dq_ds / var, shape=(nk,ns)) # was -2
        i.append(pars['kappa_color'].indexof(m.row))
        j.append(pars['sigma_kappa'].indexof(m.col))
        val.append(m.data)
        j.append(pars['kappa_color'].indexof(m.row))
        i.append(pars['sigma_kappa'].indexof(m.col))
        val.append(m.data)

        # hessian: sigma_kappa - sigma_kappa block
        dq = kappa * dq_ds / np.sqrt(var)
        # tr_wdvwdv = -4 * dq.T @ dq
        # dq = kappa**2 * dq_ds / sig
        kwdvwdvwk = dq.T @ dq
        m = sparse.coo_matrix(kwdvwdvwk, shape=(ns,ns))
        i.append(pars['sigma_kappa'].indexof(m.row))
        j.append(pars['sigma_kappa'].indexof(m.col))
        val.append(m.data)

        # encapsulate all that into the (much larger) parameter vector space
        i = np.hstack(i)
        j = np.hstack(j)
        val = np.hstack(val)
        idx = (i>=0) & (j>=0)
        n_free_pars = len(pars.free)
        hessian = sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=(n_free_pars, n_free_pars))

        return penalty, grad, hessian

    def correction(self, pars):
        """return the multiplicative correction to be applied to the model
        """
        corr = 1. + pars['kappa_color'].full[self.training_dataset.lc_data.lc_index]
        return corr

    def noise(self, pars):
        """generate one realization of the noise -- to add directly to the data
        """
        var = self.__call__(pars, jac=False)
        nn = np.random.normal(scale=np.sqrt(var), size=len(var))
        return nn[self.training_dataset.lc_data.lc_index]




import numpy as np
import pylab as pl
import sncosmo



filter_repo = '/home/nrl/salt3/naclprod/data/bandpasses/hsc-variable'


def register_bandpasses(filter_repo=filter_repo):
    """
    """
    for band in ['g', 'r', 'i', 'z', 'y', 'r2', 'i2']:
        b = sncosmo.snfitio.read_snfit_bandpass_interpolator(filter_repo, band, name=band)
        sncosmo.bandpasses._BANDPASS_INTERPOLATORS.register(b, 'hsc::' + band)
        # sncosmo.register(b, 'hsc::' + band)


def plot_transmissions(band, axes=None, title='', legend=False):
    """
    """
    if axes is None:
        fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(10,10))

    nrads = 30
    dcol = int(256 / nrads)
    for i,rad in enumerate(np.linspace(0, 28-0.1, 30)):
        b = sncosmo.get_bandpass(band, rad)
        wl = np.linspace(b.minwave(), b.maxwave(), 250)
        axes.plot(wl, b(wl), ls='-', alpha=0.25, color=pl.cm.jet_r(i * dcol),
                  label=f'rad: {rad:.2f} cm')
    axes.set_xlabel(r'$\lambda [\AA]$') #  fontsize=18)
    axes.set_ylabel(r'$T(\lambda)$') # fontsize=18)
    fig = pl.gcf()
    fig.suptitle(title)
    if legend:
        pl.legend(loc='best')


def plot_wave_eff_vs_rad(band, axes=None, xlabel='', ylabel='', title='', ls='-', label='', legend=False):
    """
    """
    if axes is None:
        fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(8,8))

    rads = np.arange(0., 28., 0.1)
    wave_eff = []
    for r in np.arange(0., 28, 0.1):
        tr = sncosmo.get_bandpass(band, r)
        wave_eff.append(tr.wave_eff)
    wave_eff = np.array(wave_eff)
    w = wave_eff[0]
    col = pl.cm.jet(int(256 * (w-4000)/(8500-3000)))

    axes.axhline(0, lw=1, color='gray')
    axes.axhspan(-0.001, 0.001, color='gray', alpha=0.1)

    axes.plot(rads, (wave_eff-wave_eff[0])/wave_eff[0], ls=ls, lw=2, color=col, label=label)
    if xlabel:
        axes.set_xlabel(xlabel)
    if ylabel:
        axes.set_ylabel(ylabel)
    pl.subplots_adjust(hspace=0.005, wspace=0.005)
    if title:
        axes.text(0.5, 0.9, f'{band}',
                  horizontalalignment='center',
                  verticalalignment='center',
                  transform = axes.transAxes)

    axes.set_ylim((-0.007, 0.007))
    if legend:
        axes.legend(loc='lower left')


def plot_throughput_vs_rad(band, axes=None, xlabel='', ylabel='', title='', ls='-', label='', legend=False):
    """
    """
    if axes is None:
        fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(8,8))

    rads = np.arange(0., 27., 0.1)
    throughputs = []
    wave_eff = []
    for r in np.arange(0., 27, 0.1):
        tr = sncosmo.get_bandpass(band, r)
        wl = np.linspace(tr.minwave(), tr.maxwave(), 100)
        throughputs.append(tr(wl).sum())
        wave_eff.append(tr.wave_eff)
    throughputs = np.array(throughputs)

    col = pl.cm.jet(int(256 * (wave_eff[0]-4000)/(8500-3000)))
    axes.axhline(0, lw=1, color='gray')
    axes.axhspan(-0.01, 0.01, color='gray', alpha=0.1)

    axes.plot(rads, (throughputs-throughputs[0])/throughputs[0], ls=ls, lw=2, color=col, label=label)
    if xlabel:
        axes.set_xlabel(xlabel)
    if ylabel:
        axes.set_ylabel(ylabel)
    pl.subplots_adjust(hspace=0.005, wspace=0.005)
    if title:
        axes.text(0.5, 0.9, f'{band}',
                  horizontalalignment='center',
                  verticalalignment='center',
                  transform = axes.transAxes)

    axes.set_ylim((-0.1, 0.1))
    if legend:
        axes.legend(loc='lower left')


def all_plots():
    """
    """
    # all bands, r,i
    fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(16,6))
    for b in ['hsc::'  + b for b in ['g', 'r', 'i', 'z', 'y']]:
        plot_transmissions(b, axes=axes, title='all bands, original r and i')

    # all bands, r2, i2
    fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(16,8))
    for b in ['hsc::'  + b for b in ['g', 'r2', 'i2', 'z', 'y']]:
        plot_transmissions(b, axes=axes, title='all bands, new r2 and i2')

    # all bands, zoom
    for b in ['hsc::'  + b for b in ['g', 'r', 'r2', 'i', 'i2', 'z', 'y']]:
        plot_transmissions(b, title=b, legend=True)

    # mean wavelength versus r
    fig, axes = pl.subplots(nrows=2, ncols=3, figsize=(12,8), sharex=True, sharey=True)
    plot_wave_eff_vs_rad('hsc::g',  axes=axes[0,0], ylabel=r'$\delta\lambda/\lambda$', title='g')
    plot_wave_eff_vs_rad('hsc::r',  axes=axes[0,1], title='r', label='r', ls=':')
    plot_wave_eff_vs_rad('hsc::r2', axes=axes[0,1], label='r2', legend=1)
    plot_wave_eff_vs_rad('hsc::i',  axes=axes[0,2], title='i', label='i', ls=':')
    plot_wave_eff_vs_rad('hsc::i2', axes=axes[0,2], xlabel='rad [cm]', label='i2', legend=1)
    plot_wave_eff_vs_rad('hsc::z',  axes=axes[1,0], xlabel='rad [cm]', ylabel=r'$\delta\lambda/\lambda$', title='hsc::z')
    plot_wave_eff_vs_rad('hsc::y',  axes=axes[1,1], xlabel='rad [cm]', title='hsc::y')
    pl.setp(axes[1,2], visible=False)
    fig.suptitle(r'HSC filter uniformity ($\lambda$)')


    # mean wavelength versus r
    fig, axes = pl.subplots(nrows=2, ncols=3, figsize=(12,8), sharex=True, sharey=True)
    plot_throughput_vs_rad('hsc::g',  axes=axes[0,0], ylabel=r'$\delta \int T/ \int T$', title='g')
    plot_throughput_vs_rad('hsc::r',  axes=axes[0,1], title='r', label='r', ls=':')
    plot_throughput_vs_rad('hsc::r2', axes=axes[0,1], label='r2', legend=1)
    plot_throughput_vs_rad('hsc::i',  axes=axes[0,2], title='i', label='i', ls=':')
    plot_throughput_vs_rad('hsc::i2', axes=axes[0,2], xlabel='rad [cm]', label='i2', legend=1)
    plot_throughput_vs_rad('hsc::z',  axes=axes[1,0], xlabel='rad [cm]', ylabel=r'$\delta\int T/\int T$', title='hsc::z')
    plot_throughput_vs_rad('hsc::y',  axes=axes[1,1], xlabel='rad [cm]', title='hsc::y')
    pl.setp(axes[1,2], visible=False)
    fig.suptitle(r'HSC filter uniformity ($\int T$)')


    return axes



    # throughput versus r



"""Utility functions and classes for sparse matrices"""

import logging
import os
import ctypes

import numpy as np
import scipy

from nacl.sparseutils_ext import _kron_product_by_line


def kron_product_by_line(a, b):
    r"""A line-wise Kronecker product of two matrices

    Parameters
    ----------
    a : (k, n) sparse matrix
        first matrix of the product, must be sparse preferably in CSR format.
        The matrix is converted to CSR format if needed.
    b : (k, m) dense matrix
        second matrix of the product

    Returns
    -------
    c : (k, m*n) sparse matrix in COO format
        Line-wise Kronecker product of `a` and `b`

    Examples
    --------
    >>> import numpy as np
    >>> import scipy as sp
    >>> from nacl.sparseutils import kron_product_by_line
    >>> a = sp.sparse.csr_matrix(np.array([[0, 2], [5, 0]]))
    >>> b = np.array([[1, 2], [3, 4]])
    >>> kron_product_by_line(a, b).toarray()
    array([[ 0.,  0.,  2.,  4.],
           [15., 20.,  0.,  0.]])

    """
    if not (a.ndim == 2 and b.ndim == 2):
        raise ValueError("The both arrays should be 2-dimensional.")

    if not a.shape[0] == b.shape[0]:
        raise ValueError(
            "The number of rows for both arrays should be equal.")

    if not scipy.sparse.issparse(a):
        raise ValueError("`a` must be a sparse array.")

    if scipy.sparse.issparse(b):
        raise ValueError("`b` must be a dense array.")

    a = a.tocsr()

    # estimate the number of non-zero triplets in result
    non_zeros = a.indptr[1:] - a.indptr[:-1]
    estimated_result_size = (non_zeros * b.shape[1]).sum()

    rows = np.zeros(estimated_result_size, dtype=np.int32)
    cols = np.zeros(estimated_result_size, dtype=np.int32)
    vals = np.zeros(estimated_result_size, dtype=np.float64)

    n = _kron_product_by_line(
        a.shape[0], a.shape[1], b.shape[1], non_zeros.max(),
        a.indices, a.indptr, a.data,
        b,
        rows, cols, vals)

    return scipy.sparse.coo_matrix(
        (vals[:n], (rows[:n], cols[:n])),
        shape=(a.shape[0], a.shape[1] * b.shape[1]))


class CooMatrixBuff:
    def __init__(self, shape, estimated_nnz, increment=1.3):
        self.shape = shape
        self.size = estimated_nnz
        self.increment = increment
        self.i = np.zeros(self.size).astype(np.int64)
        self.j = np.zeros(self.size).astype(np.int64)
        self.val = np.zeros(self.size)
        self.ptr = 0

    def _resize(self):
        logging.warning(
            'need to resize CooMatrixBuff: revise your estimates of non-zero terms !')
        new_size = self.size * self.increment
        self.i = np.resize(self.i, new_size)
        self.j = np.resize(self.j, new_size)
        self.val = np.resize(self.val, new_size)
        self.i[self.ptr:] = 0
        self.j[self.ptr:] = 0
        self.val[self.ptr:] = 0.
        self.size = new_size

    def append(self, i, j, val, free_pars_only=False):
        """
        """
        logging.info(f'appending {len(i)} to buffer at location {self.ptr}')
        sz = len(i)
        assert (len(j) == sz) and (len(val) == sz)

        if (self.ptr + sz) > self.size:
            self._resize()

        # if free_pars_only:
        #     sz = lib.append(
        #         len(i), self.ptr,
        #         i.astype(np.int64), j.astype(np.int64), val,
        #         self.i, self.j, self.val,
        #         1)
        #     self.ptr += sz
        # else:
        #     sz = lib.append(
        #         len(i), self.ptr,
        #         i.astype(np.int64), j.astype(np.int64), val,
        #         self.i, self.j, self.val,
        #         0)
        #     self.ptr += sz

        if free_pars_only:
            idx = j >= 0
            sz = idx.sum()
            self.i[self.ptr:self.ptr+sz] = i[idx]
            self.j[self.ptr:self.ptr+sz] = j[idx]
            self.val[self.ptr:self.ptr+sz] = val[idx]
            self.ptr += sz
        else:
            self.i[self.ptr:self.ptr+sz] = i
            self.j[self.ptr:self.ptr+sz] = j
            self.val[self.ptr:self.ptr+sz] = val
            self.ptr += sz
        logging.info('done')

    def tocoo(self):
        return scipy.sparse.coo_matrix(
            (self.val[:self.ptr], (self.i[:self.ptr], self.j[:self.ptr])),
            self.shape)


class CooMatrixBuff2:
    def __init__(self, shape, increment=1.3):
        self.shape = shape
        self.increment = increment
        self._row = []
        self._col = []
        self._data = []
        self._idx = []

    def append(self, i, j, val):
        """
        """
        self._row.append(i)
        self._col.append(j)
        self._data.append(val)
        self._idx.append(j>=0)

    def tocoo(self):
        idx = np.hstack(self._idx)
        r = scipy.sparse.coo_matrix(
            (np.hstack(self._data)[idx],
             (np.hstack(self._row)[idx], np.hstack(self._col)[idx])),
            self.shape)
        return r

    @property
    def row(self):
        return self._row

    @property
    def col(self):
        return self._col

    @property
    def data(self):
        return self._data



"""SN light curve eval unit
"""

import logging

import time
import numpy as np
import scipy

# import numexpr as ne

try:
    from sparse_dot_mkl import gram_matrix_mkl, dot_product_mkl
except:
    logging.warning('module: `sparse_dot_mkl` not available')
else:
    logging.info('sparse_dot_mkl found. Building hessian should be faster.')

from nacl.sparseutils import kron_product_by_line, CooMatrixBuff, CooMatrixBuff2
from bbf import magsys


class LightcurveEvalUnit:
    """The evaluation unit for light curves.

    This class evaluates all light curves present in the training dataset at
    once for increased efficiency.

    """
    def __init__(self, model):
        """
        Initialize the LightcurveEvalUnit.

        Parameters
        ----------
        model : object
            The main model object containing the training dataset.
        """
        self.model = model
        self.training_dataset = model.training_dataset
        self.gram = self.model.gram.todense()
        self.color_law = self.model.color_law

        nb_lightcurves = len(self.training_dataset.lc_db)
        filter_db_basis_size = len(model.filter_db.basis)
        F = np.zeros((nb_lightcurves, filter_db_basis_size))
        for lc in self.training_dataset.lc_db:
            # tqz, _ = self.model.filter_db.insert(lc.band, z=lc.z)
            tr_data = self.model.filter_db.insert(lc.band, z=lc.z)
            F[lc.lc_index, :] = tr_data.tq

        self.flux_scales = self.compute_flux_scales()
        self.flux_scales *= self.model.norm

        # filter projections
        self.filter_projections = (self.gram @ F.T).T
        self.meas_filter_projections = \
            self.filter_projections[self.training_dataset.lc_data.lc_index]

    def compute_flux_scales(self):
        """Compute the flux scales for the training dataset.

        The flux scales are necessary to convert model fluxes to observed
        fluxes in units defined by the observer.

        Returns
        -------
        numpy.array
            Array of flux scales.

        Notes
        -----
        Here, we assume that all mags are in the AB mag system. This may not be true.
        This need to be corrected.

        """
        zp_scale = self.training_dataset.lc_data.zp_scale

        # also insert the filter at z = 0, needed to compute integrals of AB spectrum
        bands = np.unique(self.training_dataset.lc_db.band)
        for b in bands:
            self.model.filter_db.insert(b, z=0.)

        ms = magsys.SNMagSys(self.model.filter_db)
        int_AB_dict = dict(zip(bands, [10.**(0.4 * ms.get_zp(b.lower())) for b in bands]))
        int_AB = np.array([int_AB_dict[b] for b in self.training_dataset.lc_data.band])
        flux_scales = int_AB / zp_scale
        return flux_scales

    def __call__(self, pars, jac=False):
        """Evaluate the light curve model for a given set of parameters.

        Parameters
        ----------
        pars : nacl.fitparameters.FitParameters
            The current set of fit parameters.
        jac : bool, optional
            Whether to compute and return the Jacobian matrix. Defaults to False.

        Returns
        -------
        numpy.array
            Model flux values.
        scipy.sparse.coo_matrix, optional
            Jacobian matrix if `jac` is True.

        """
        lc_data = self.training_dataset.lc_data
        lc_db = self.training_dataset.lc_db

        wl_basis = self.model.basis.bx
        ph_basis = self.model.basis.by
        n_wl, n_ph = len(wl_basis), len(ph_basis)

        # pick matrix
        n_lc, n_meas = len(lc_db), len(lc_data)

        # phases
        zz = 1. + lc_data.z
        tmax = pars['tmax'].full[lc_data.sn_index]
        restframe_phases = (lc_data.mjd - tmax) / zz
        J_phase_sparse = ph_basis.eval(restframe_phases + self.model.delta_phase)
        J_phase = np.array(J_phase_sparse.todense())


        if 'eta_calib' in pars._struct.slices:
            calib_corr = 1. + pars['eta_calib'].full[lc_data.band_index]
        else:
            calib_corr = np.ones(len(lc_data.band_index))

        if 'kappa_color' in pars._struct.slices:
            cs_corr = 1. + pars['kappa_color'].full[lc_data.lc_index]
        else:
            cs_corr = np.ones(len(lc_data.lc_index))

        M0 = pars['M0'].full.reshape(n_ph, n_wl)
        M1 = pars['M1'].full.reshape(n_ph, n_wl)
        C0_ = np.array(M0.dot(self.meas_filter_projections.T))
        C0 = (J_phase * C0_.T).sum(axis=1)
        C1_ = np.array(M1.dot(self.meas_filter_projections.T))
        C1 = (J_phase * C1_.T).sum(axis=1)

        X0 = pars['X0'].full[lc_data.sn_index]
        X1 = pars['X1'].full[lc_data.sn_index]
        col = pars['c'].full[lc_data.sn_index]

        # color law - so here, we decide to move it out of the integral
        # maybe we need to add a small correction
        cl_pars = pars['CL'].full

        restframe_wavelength = lc_data.wavelength / zz
        cl_pol, J_cl_pol = self.color_law(restframe_wavelength,
                                          cl_pars, jac=jac)
        cl = np.power(10., 0.4 * col * cl_pol)

        pca = C0 + X1 * C1
        model_val = X0 * pca * cl * zz * calib_corr * cs_corr

        if not jac:
            v = np.zeros(len(self.training_dataset))
            v[lc_data.row] = self.flux_scales * model_val
            return v

        # jacobian
        N = len(self.training_dataset)
        n_free_pars = len(pars.free)

        # since the hstack is taking a lot of time and memory, we do things differently:
        # we allocate 3 large buffers for the jacobian i, j, vals, and we
        # update them in place.

        # estimated size of the derivatives
        logging.debug(' ... kron')
        K = kron_product_by_line(J_phase_sparse, self.meas_filter_projections)
        logging.debug(f'     -> done, K.nnz={K.nnz} nnz_real={(K.data != 0.).sum()} {len(K.row)}')

        estimated_size = 2 * K.nnz   # dMdM0 and dMdM1
        estimated_size += 6 * N      # local parameters (dMdX0, dMdX1, dMdcol, dMtmax, dMdeta, dMdkappa)
        nnz = len(J_cl_pol.nonzero()[0])
        estimated_size += nnz
        logging.debug(f'estimated size: {estimated_size}')

        buff = CooMatrixBuff2((N, n_free_pars)) # , estimated_size)
        self.K = K
        self.X0 = X0
        self.cl = cl
        self.calib_corr = calib_corr
        self.cs_corr = cs_corr
        self.zz = zz

        # we start with the largest derivatives: dMdM0 and dMdM1
        # dMdM0
        # we could write it as:
        # v_ = X0[K.row] * K.data * cl[K.row] * calib_corr[K.row] * cs_corr[K.row] * zz[K.row]
        # but it is slow. So, we re-arrange it as:
        i_ = lc_data.row[K.row]
        v_ = X0 * cl * calib_corr * cs_corr * zz
        # vv_, dd_ = v_[K.row], K.data
        # v_ = ne.evaluate('vv_ * dd_')
        v_ = v_[K.row] * K.data
        buff.append(i_,
                    pars['M0'].indexof(K.col),
                    v_)

        # dMdM1
        # X1_ = X1[K.row]
        buff.append(lc_data.row[K.row],
                    pars['M1'].indexof(K.col),
                    v_ * X1[K.row])

        del K
        del i_
        del v_

        # dMdtmax
        phase_basis = self.model.basis.by
        dJ = np.array(phase_basis.deriv(restframe_phases + self.model.delta_phase).todense())
        dC0 = (dJ * C0_.T).sum(axis=1)
        dC1 = (dJ * C1_.T).sum(axis=1)
        buff.append(lc_data.row,
                    pars['tmax'].indexof(lc_data.sn_index),
                    # ne.evaluate('-X0 * (dC0 + X1 * dC1) * cl * calib_corr * cs_corr'))
                    -X0 * (dC0 + X1 * dC1) * cl * calib_corr * cs_corr)

        del dJ
        del C0_
        del C1_

        # dMdcl
        JJ = scipy.sparse.coo_matrix(J_cl_pol)
        model_val_, col_, d_ = model_val[JJ.row], col[JJ.row], JJ.data
        buff.append(JJ.row,
                    pars['CL'].indexof(JJ.col),
                    # ne.evaluate('model_val_ * 0.4 * np.log(10.) * col_ * d_'))
                    # self.flux_scales[JJ.row] * model_val[JJ.row] * 0.4 * np.log(10.) * col[JJ.row] * JJ.data)
                    model_val[JJ.row] * 0.4 * np.log(10.) * col[JJ.row] * JJ.data)

        del JJ
        del model_val_
        del col_
        del d_

        # dMdX0
        buff.append(lc_data.row,
                    pars['X0'].indexof(lc_data.sn_index),
                    pca * cl * zz * calib_corr * cs_corr)

        # dMdX1
        buff.append(lc_data.row,
                    pars['X1'].indexof(lc_data.sn_index),
                    X0 * C1 * cl * zz * calib_corr * cs_corr)

        # dMdcol
        buff.append(
            lc_data.row,
            pars['c'].indexof(lc_data.sn_index),
            model_val * 0.4 * np.log(10.) * cl_pol)

        # dMdeta
        if 'eta_calib' in pars._struct:
            buff.append(lc_data.row,
                        pars['eta_calib'].indexof(lc_data.band_index),
                        X0 * pca * cl * zz * cs_corr)

        # dMkappa
        if 'kappa_color' in pars._struct:
            buff.append(lc_data.row,
                        pars['kappa_color'].indexof(lc_data.band_index),
                        X0 * pca * cl * zz * calib_corr)

        logging.debug(' -> tocoo()')
        J = buff.tocoo()
        del buff

        # multiply the data by the flux scales
        # to express fluxes in observer units
        J.data *= self.flux_scales[J.row]
        v = np.zeros(len(self.training_dataset))
        v[lc_data.row] = self.flux_scales * model_val

        return v, J



import numpy as np
from nacl.dataset import TrainingDataset
from nacl.models import salt2
from lemaitre import bandpasses
import logging


def test_model():
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s %(message)s')
    filterlib = bandpasses.get_filterlib()

    # Load the training data set
    tds = TrainingDataset.read_parquet("data/test_datasets/test_datasets_blind.parquet", filterlib=filterlib)

    model = salt2.get_model(tds)
    pars = model.init_pars()

    # to test the evaluation of the model and test the presence of inf, nan...
    v = model(pars)
    assert v is not None
    # assert the values of v are the ones we expect

    # to test the model derivatives
    v, J = model(pars, jac=True)
    assert v is not None
    assert J is not None
    # assert the values of v and J correspond to what we expect


def test_model_cache():
    """Test the caching system implemented in the SALT2Like model
    """
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s %(message)s')
    filterlib = bandpasses.get_filterlib()

    # Load the training data set
    tds = TrainingDataset.read_parquet("data/test_datasets/test_datasets_blind.parquet", filterlib=filterlib)

    # if we disable de cache, make sure that everything
    # is indeed recomputed
    model = salt2.get_model(tds, disable_cache=True)
    pars = model.init_pars()
    v = model(pars)
    assert model.cached is False
    vv = model(pars)
    assert model.cached is False
    assert np.allclose(v, vv)
    assert id(v) != id(vv)

    v, J = model(pars, jac=1)
    assert model.cached is False
    vv, JJ = model(pars, jac=1)
    assert model.cached is False
    assert np.allclose(v, vv)
    # what is super strange, is that J and JJ may be ordered differently
    D = J.todense() - JJ.todense()
    assert np.count_nonzero(D) == 0
    #    assert np.allclose(J.data, JJ.data)
    #    assert np.allclose(J.row, JJ.row)
    #    assert np.allclose(J.col, JJ.col)
    assert id(v) != id(vv)
    assert id(J) != id(JJ)

    # now, if we enable to cache, make sure that cache system works
    model = salt2.get_model(tds, disable_cache=False)
    pars = model.init_pars()
    v = model(pars)
    assert model.cached is False
    vv = model(pars)
    assert model.cached is True
    assert np.allclose(v, vv)
    assert id(v) == id(vv)

    # this time we want the jacobian
    # which, at this point, is not stored in the cache
    v, J = model(pars, jac=1)
    assert model.cached is False
    vv, JJ = model(pars, jac=1)
    assert model.cached is True
    assert np.allclose(v, vv)
    assert np.allclose(J.data, JJ.data)
    assert id(v) == id(vv)

    # change the parameter vector a little
    pars.full[222] += 1.E-6
    vvv = model(pars)
    assert model.cached is False



import numpy as np
import pylab as pl
import sncosmo
from nacl.models.salt2.salt import SALT2Like
source = sncosmo.get_source('salt2', version='2.4')
import matplotlib.colors as mcolors
import numpy.polynomial.polynomial as nppol

def plot_stacked_residuals_snake(tds, pars, v, var):
    """
    Plotting the stacked residuals comparing the all of the data and the model
    
    Parameters:
    ----------
    tds : nacl.dataset.TrainingDataset 
          initial training dataset
          
    model : nacl.models.salt2.salt.SALT2Like
            the model
        
    v : numpy.array
        result values that come out of the NaCl model. Default=None
    
    """

    
    wl = tds.spec_data.wavelength / (1. + tds.spec_data.z)
    ph = (tds.spec_data.mjd - pars['tmax'].full[tds.spec_data.sn_index]) / (1 + tds.spec_data.z)

    pl.figure(figsize=(14, 10))
    idx = tds.spec_data.valid > 0
    res = tds.spec_data.flux - v[len(tds.lc_data.flux):len(tds.lc_data.flux)+len(tds.spec_data.flux)]
    wres = res / (tds.spec_data.fluxerr + np.sqrt(var[len(tds.lc_data.flux):len(tds.lc_data.flux)+len(tds.spec_data.flux)]))
    ii = ph[idx] < 50.
    pl.hexbin(wl[idx][ii], ph[idx][ii], wres[idx][ii], vmin=-2.5, vmax=2.5)
    pl.title('weighted residuals Spectra', fontsize=16)
    pl.xlabel(r'restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()
    pl.show()
    
    wl = tds.lc_data.wavelength / (1. + tds.lc_data.z)
    ph = (tds.lc_data.mjd - pars['tmax'].full[tds.lc_data.sn_index]) / (1 + tds.lc_data.z)

    pl.figure(figsize=(14, 10))
    idx = tds.lc_data.valid > 0
    res = tds.lc_data.flux - v[:len(tds.lc_data.flux)]
    wres = res / (tds.lc_data.fluxerr + np.sqrt(var[:len(tds.lc_data.flux)]) )
    ii = ph[idx] < 50.
    pl.hexbin(wl[idx][ii], ph[idx][ii], wres[idx][ii], vmin=-2.5, vmax=2.5)
    pl.title('weighted residuals LCs', fontsize=16)
    pl.xlabel(r'restframe $\lambda [\AA]$', fontsize=16)
    pl.ylabel('restframe phase [days]', fontsize=16)
    pl.colorbar()
    pl.show()
    
def plot_band_residuals_z(tds, v, var):
    """
    Plotting residuals in each band
    
    Parameters:
    -----------
    tds : nacl.TrainingDataset
        The training dataset
    v : np.array
        The model evaluation
    var : np.array
        The error model
    """
    nb_bands = len(tds.lc_data.band_set)
    v_lc = v[:len(tds.lc_data)]
    
    fig, axs = pl.subplots(nrows = nb_bands, ncols=2, width_ratios=np.array([3,1]), sharey='row', figsize=(14,10))
    bands = np.unique(tds.lc_data.nt['band'])
    color = ['azure', 'black', 'blue', 'brown', 'chocolate', 'cyan', 'darkblue', 'darkcyan', 'darkgreen', 'darkolivegreen', 'darkred', 'darkorange',
    'darkmagenta', 'darkviolet', 'green', 'olive', 'red', 'orange', 'magenta', 'lime', 'gold', 'lavender', 'pink', 'teal']
    #cc = np.random.choice(np.array(list(mcolors.cnames.keys())), size=nb_bands, replace=False)
    cc = np.random.choice(np.array(color), size=nb_bands, replace=False)
    for i in range(nb_bands):
        idx = tds.lc_data.band == bands[i]
        band_name = tds.lc_data.band[idx][0]
        
        res = tds.lc_data.flux[idx] - v_lc[idx]
        z = tds.lc_data.z[idx]
        
        axs[i,0].errorbar(z, res, yerr = tds.lc_data.fluxerr[idx], fmt='.', label=band_name, color=cc[i])
        axs[i,1].hist(res, bins=100, orientation='horizontal', color=cc[i])
        axs[i,1].set_ylim(np.mean(res) -5*np.std(res), np.mean(res) + 5*np.std(res))
        
        #reindexing = sorted(enumerate(z), key = lambda x:x[1])
        #z_idx = np.array([index for index, value in reindexing])
        #axs[i].fill_between(z[z_idx], res[z_idx]+np.sqrt(var[:len(tds.lc_data)][idx][z_idx]), res[z_idx]-np.sqrt(var[:len(tds.lc_data)][idx][z_idx]), alpha=0.5)
        
        
        axs[i,0].set_xlabel('z')
        axs[i,0].set_ylabel('res')
        axs[i,0].legend()
        
    pl.tight_layout()
    
    
def plot_band_residuals_phase(tds, pars, v, var):
    """
    Plotting residuals in each band
    
    Parameters:
    -----------
    tds : nacl.TrainingDataset
        The training dataset
    v : np.array
        The model evaluation
    var : np.array
        The error model
    """
    nb_bands = len(tds.lc_data.band_set)
    v_lc = v[:len(tds.lc_data)]
    
    fig, axs = pl.subplots(nrows = nb_bands, ncols=2, width_ratios=np.array([3,1]), sharey='row', figsize=(14,10))
    bands = np.unique(tds.lc_data.nt['band'])
    color = ['azure', 'black', 'blue', 'brown', 'chocolate', 'cyan', 'darkblue', 'darkcyan', 'darkgreen', 'darkolivegreen', 'darkred', 'darkorange',
    'darkmagenta', 'darkviolet', 'green', 'olive', 'red', 'orange', 'magenta', 'lime', 'gold', 'lavender', 'pink', 'teal']
    #cc = np.random.choice(np.array(list(mcolors.cnames.keys())), size=nb_bands, replace=False)
    cc = np.random.choice(np.array(color), size=nb_bands, replace=False)
    for i in range(nb_bands):
        idx = tds.lc_data.band == bands[i]
        band_name = tds.lc_data.band[idx][0]
        
        res = tds.lc_data.flux[idx] - v_lc[idx]
        p = tds.lc_data.mjd[idx] - pars['tmax'].full[tds.lc_data.sn_index][idx]
        
        axs[i,0].errorbar(p, res, yerr = tds.lc_data.fluxerr[idx], fmt='.', label=band_name, color=cc[i])
        axs[i,1].hist(res, bins=100, orientation='horizontal', color=cc[i])
        axs[i,1].set_ylim(np.mean(res) -5*np.std(res), np.mean(res) + 5*np.std(res))
        
        #reindexing = sorted(enumerate(z), key = lambda x:x[1])
        #z_idx = np.array([index for index, value in reindexing])
        #axs[i].fill_between(z[z_idx], res[z_idx]+np.sqrt(var[:len(tds.lc_data)][idx][z_idx]), res[z_idx]-np.sqrt(var[:len(tds.lc_data)][idx][z_idx]), alpha=0.5)
        
        
        axs[i,0].set_xlabel('phase')
        axs[i,0].set_ylabel('res')
        axs[i,0].legend()
        
    pl.tight_layout()
    
    
def plot_cs_variance(tds, pars, cs):
    """
    Color scatter varaince
    """
    lc_data = tds.lc_data
    _restframe_wl = lc_data.wavelength / (1. + lc_data.z)
    _, ii = np.unique(lc_data.lc_index, return_index=True)
    ii.sort()
    restframe_wl = _restframe_wl[ii]
    
    sigma_kappa = pars['sigma_kappa'].full
    #var = np.exp(np.polyval(sigma_kappa, cs.reduced_restframe_wl))
    var = np.exp(nppol.Polynomial(sigma_kappa)(cs.reduced_restframe_wl))
    
    pl.figure(figsize=(12,8))
    pl.plot(restframe_wl, var, 'b.')
    pl.xlabel('restframe wavelengths')
    pl.ylabel(r'$\sigma_{\kappa}$')
    pl.title('color scatter variance')
    
def plot_cs_cl(tds, logs, cl):
    """
    Color law after each fit
    """
    lc_data = tds.lc_data
    _restframe_wl = lc_data.wavelength / (1. + lc_data.z)
    _, ii = np.unique(lc_data.lc_index, return_index=True)
    ii.sort()
    restframe_wl = _restframe_wl[ii]
    
    pars_final = logs[-1].pars['CL'].full
    pars_no_cs = logs[-2].pars['CL'].full
    pars_init = logs[0].pars['CL'].full
    
    fig, axs = pl.subplots(nrows = 2, ncols=1, height_ratios=np.array([3,1]), figsize=(14,10))
    
    axs[0].plot(restframe_wl, cl(restframe_wl, pars_init)[0], 'b.', label='SALT2.4')
    axs[0].plot(restframe_wl, cl(restframe_wl, pars_no_cs)[0], 'r.', label='no color scatter')
    axs[0].plot(restframe_wl, cl(restframe_wl, pars_final)[0], 'g.', label='full fit')
    
    axs[0].set_xlabel('restframe wavelengths')
    axs[0].set_label(r'CL$(\lambda)$')
    axs[0].set_title('Color Law evolution')
    
    axs[0].legend()
    
    axs[1].plot(restframe_wl, cl(restframe_wl, pars_init)[0] - cl(restframe_wl, pars_final)[0], 'b.', label='SALT2.4 - full fit')
    
    axs[1].set_xlabel('restframe wavelengths')
    axs[1].set_label(r'$\Delta$CL$(\lambda)$')
    axs[1].set_title('Color Law difference')
    
    axs[1].legend()
    
    
def plot_kappa(tds, pars):
    """
    Color scatter coeffs
    """
    lc_data = tds.lc_data
    _restframe_wl = lc_data.wavelength / (1. + lc_data.z)
    _, ii = np.unique(lc_data.lc_index, return_index=True)
    ii.sort()
    restframe_wl = _restframe_wl[ii]
    
    pl.figure(figsize=(12,8))
    pl.plot(restframe_wl, pars['kappa_color'].full, 'b.')
    pl.xlabel('restframe wavelengths')
    pl.ylabel(r'$\kappa_{color}$')
    pl.title('kappa color')
    




from nacl.dataset import TrainingDataset
from nacl.models import salt2
from nacl.train import TrainSALT2Like
from lemaitre import bandpasses

import pytest


def test_training(tmp_path):
    filterlib = bandpasses.get_filterlib()

    # Load the training data set
    tds = TrainingDataset.read_parquet(
        "data/test_datasets/test_datasets_blind.parquet",
        filterlib=filterlib)

    model = salt2.get_model(tds)
    # tds.plot_sample()
    # plt.show()

    pars = model.init_pars()
    # to test the evaluation of the model and test the presence of inf, nan...
    v = model(pars)
    assert v is not None

    # Initialisation of the trainer
    trainer = TrainSALT2Like(tds, variance_model='simple_snake')
    # trainer.plot_lc("ZTF17aadlxmv", numfit=None, plot_variance=True)
    # plt.show()

    # # Training
    # path = str(tmp_path / "lemaitre_nacl_tests")
    # trainer.train_salt2_model(save=True, path=path)

    # # Outputs
    # pars_trained = trainer.log[-1].pars
    # v_trained = trainer.log[-1].v  # evaluated model
    # # cov_matrix = trainer.log[-1].minz.get_cov_matrix()[0]

    # # Tests
    # assert pars_trained is not None
    # assert v_trained is not None
